//
// "$Id: DirectFB.H,v 1.3 2006-10-31 09:31:14 nikego Exp $"
//
// DirectFB header file for the Fast Light Tool Kit (FLTK).
//
// Copyright 1998-2005 by Bill Spitzak and others.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Library General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Library General Public License for more details.
//
// You should have received a copy of the GNU Library General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA.
//
// Please report all bugs and problems on the following page:
//
//     http://www.fltk.org/str.php
//

// Do not directly include this file, instead use <FL/x.H>.  It will
// include this file if DIRECTFB is defined.  This is to encourage
// portability of even the system-specific code...

#ifndef Fl_X_H
#  error "Never use <DirectFB.H> directly; include <FL/x.H> instead."
#endif // !Fl_X_H

#include <directfb.h>
#include <directfb_version.h>

//////////////////missing in Sharp DirectFB.
#if DIRECTFB_MAJOR_VERSION == 0
typedef struct {
     int        magic;

     DFBRegion *regions;
     int        max_regions;
     int        num_regions;

     DFBRegion  bounding;
} DFBUpdates;


void dfb_updates_init( DFBUpdates      *updates,
                       DFBRegion       *regions,
                       int              max_regions );

void dfb_updates_add ( DFBUpdates      *updates,
                       const DFBRegion *region );

void dfb_updates_stat( DFBUpdates      *updates,
                       int             *ret_total,
                       int             *ret_bounding );


static inline void
dfb_updates_reset( DFBUpdates *updates )
{
     //D_MAGIC_ASSERT( updates, DFBUpdates );

     updates->num_regions = 0;
}

typedef unsigned char u8;
typedef unsigned short u16;
typedef unsigned int u32;

typedef unsigned short uint16_t;
typedef unsigned int uintptr_t;
#endif
//////////////////

#include <config.h>
#include <FL/xutils.h>
#include <FL/FL_DFB_Window.h>
#include <FL/FL_DFB_WindowManager.h>


// In some of the distributions, the gcc header files are missing some stuff:
#ifndef LPMINMAXINFO
#define LPMINMAXINFO MINMAXINFO*
#endif
#ifndef VK_LWIN
#define VK_LWIN 0x5B
#define VK_RWIN 0x5C
#define VK_APPS 0x5D
#endif

#include <stdio.h>

// This is the super interface, it's the entry point to all functionality.
extern IDirectFB*               fl_dfb; 
extern IDirectFBDisplayLayer*   fl_display_layer;
extern GC                       fl_screen; //The primary surface, i.e. the "screen".

#if 0
#  define _TRACE(p...) printf(p)
#  define  TRACE(p...) printf(p)
#else
#  define _TRACE(p...)
#  define __TRACE(p...) printf(p)
#  define TRACE(p...)
#endif

void print_last_dump();

#ifdef DEBUG 

#define DFBCHECK(x...)\
  do {\
    DFBResult err = x;\
    if (err != DFB_OK){\
        fprintf( stderr, "Error '%s' occured in %s <%d>:\n\t", DirectFBErrorString(err), __FILE__, __LINE__ );\
        fprintf( stderr, #x );\
        print_last_dump();\
       /* DirectFBErrorFatal( #x, err );*/\
      }\
  } while(0)
  
#else

    #define DFBCHECK(x...) x
    
#endif
  
/******************************************************/
/*   DirectFB's  assist functions                     */
/******************************************************/

static inline int
get_caps( GC gc ){
  DFBSurfaceCapabilities ret_caps = DSCAPS_NONE;
  if (gc && gc->surface)
    DFBCHECK(gc->surface->GetCapabilities(gc->surface, &ret_caps));
  return ret_caps;
}   

static inline int
get_pixel_format(GC gc){
  DFBSurfacePixelFormat pixel_format = DSPF_UNKNOWN;
  if (gc && gc->surface)
    gc->surface->GetPixelFormat(gc->surface, &pixel_format);
  return (int)pixel_format;
}

static inline void
get_size(GC gc, int* width, int* height ){
  if (gc && gc->surface)
    DFBCHECK(gc->surface->GetSize(gc->surface, width, height));
}

static inline void
get_visible_rectangle( GC gc, DFBRectangle* rect ){
  if (gc && gc->surface)
    DFBCHECK( gc->surface->GetVisibleRectangle( gc->surface, rect ) );
}

// summary
static inline int
get_surface_description(GC gc, DFBSurfaceDescription* desc){
  if (gc && gc->surface) {
    desc->flags     = (DFBSurfaceDescriptionFlags)(DSDESC_CAPS|DSDESC_WIDTH|DSDESC_HEIGHT|DSDESC_PIXELFORMAT);    
    desc->caps      = (DFBSurfaceCapabilities)get_caps( gc ); 
    get_size( gc, &desc->width , &desc->height );
    desc->pixelformat = (DFBSurfacePixelFormat)get_pixel_format( gc );
    return DFB_OK;
  }
  return DFB_INVARG;
}
    
static inline void
get_palette(GC gc, IDirectFBPalette** pal){
  if (gc && gc->surface)
    DFBCHECK( gc->surface->GetPalette( gc->surface, pal ) );
}

static inline void* 
lock_surface(GC gc, int flags, int *ret_pitch){
  void* buffer = 0;
  if (gc && gc->surface)
    if (DFB_OK == gc->surface->Lock(gc->surface, (DFBSurfaceLockFlags)flags,\
      &buffer, ret_pitch))
      return buffer;
  return 0;   
}

static inline void 
unlock_surface(GC gc){
  if (gc && gc->surface)
    gc->surface->Unlock( gc->surface ); 
}

static inline void
flip_gc(GC gc, DFBRegion *region, int flags){
#if 0    
    DFBCHECK(gc->surface->Flip( gc->surface, region,( DFBSurfaceFlipFlags)flags ) );
#endif    
}
    
static inline void
clear_gc(GC gc, uchar red, uchar green, uchar blue, uchar alpha){
  if (gc && gc->surface)
    DFBCHECK(gc->surface->Clear(gc->surface, red, green, blue, alpha));
}
    
static inline int
set_clip(GC gc, DFBRegion* reg){
  if (gc && gc->surface){
    if( reg ){
      DFBRegion r = *reg;
      if (r.x2) r.x2--;
      if (r.y2) r.y2--;
      return gc->surface->SetClip( gc->surface, &r );
    }else
      return gc->surface->SetClip( gc->surface, reg );
  }
  return DFB_INVARG;
}

static inline int
get_clip(GC gc, DFBRegion* reg){
  if (gc && gc->surface)
    if (reg){
      return gc->surface->GetClip(gc->surface, reg);
    }
  return DFB_INVARG;
}

inline void
set_drawing_flags(GC gc, int flags){
  if (gc && gc->surface)
    DFBCHECK(gc->surface->SetDrawingFlags(gc->surface, (DFBSurfaceDrawingFlags)flags));
}

static inline void
set_src_blend_function( GC gc, int func ){
  if (gc && gc->surface)
   gc->surface->SetSrcBlendFunction( gc->surface, (DFBSurfaceBlendFunction) func );
}

static inline void
set_dst_blend_function( GC gc, int func ){
  if (gc && gc->surface)
   gc->surface->SetDstBlendFunction( gc->surface, (DFBSurfaceBlendFunction) func );
}

static inline void
set_porter_duff(GC gc, int rule){
  if (gc && gc->surface)
    gc->surface->SetPorterDuff( gc->surface, (DFBSurfacePorterDuffRule)rule );
}

static inline void
set_blitting_flags(GC gc, int flags){
    DFBCHECK( gc->surface->SetBlittingFlags(gc->surface, (DFBSurfaceBlittingFlags)flags ) );
}

static inline void
blit(GC dest,GC src, DFBRectangle* rect, int x , int y){
  if (src && dest && dest->surface && src->surface)
    DFBCHECK(dest->surface->Blit(dest->surface,src->surface, rect, x, y));
}
    
static inline void
set_surface_font(GC gc, IDirectFBFont* font){
  if (gc && gc->surface)
    DFBCHECK(gc->surface->SetFont(gc->surface, font));  
}

static inline IDirectFBFont* 
get_surface_font(GC gc){
    IDirectFBFont* ret_font = 0;
    if (gc && gc->surface)
      gc->surface->GetFont(gc->surface, &ret_font);
    return ret_font;
}

static inline int
set_color(GC gc, u8 _r,u8 _g,u8 _b,u8 _a){
  if (gc && gc->surface)
  {
    gc->r = _r;
    gc->g = _g;
    gc->b = _b;
    gc->a = _a;
    DFBCHECK( gc->surface->SetColor( gc->surface,_r,_g,_b,_a ) );
  }
  return 0;
}

/*static inline int
set_color(GC gc, u8 _color){
  if (gc && gc->surface)
    DFBCHECK( gc->surface->SetColorIndex(gc->surface, _color));
  return 0;
}*/  

static inline int
set_src_color_key(GC gc, u8 _color){
  if (gc && gc->surface)
    DFBCHECK(gc->surface->SetSrcColorKeyIndex(gc->surface, _color));
  return 0;
}

static inline int
set_src_color_key(GC gc, u8 r, u8 g, u8 b){
  if (gc && gc->surface)
    DFBCHECK(gc->surface->SetSrcColorKey( gc->surface, r, g, b));
  return 0;
}

static inline int
set_dst_color_key( GC gc, u8 _color ){
    DFBCHECK(gc->surface->SetDstColorKeyIndex(gc->surface, _color ));
    return 0;
}

static inline int
set_dst_color_key(GC gc, u8 r, u8 g , u8 b ){
  if (gc && gc->surface)
    DFBCHECK(gc->surface->SetDstColorKey(gc->surface, r, g, b));
  return 0;
}

void
draw_lines(GC gc, Region reg, int n );

void
draw_string(GC gc, const char* str, int n,  int x, int y, DFBSurfaceTextFlags flags );

void
get_string_width(IDirectFBFont *font, const char* s, int n, int * width);

int
set_clip_mask(Pixmap mask);

/*********************************************************************
 * window's functions
 * */
#if 0
static inline DFBWindowID
get_id( IDirectFBWindow* window ){
    DFBWindowID ret_id;
    if ( window && ( DFB_OK == window->GetID( window, &ret_id ) ) )
        return ret_id;
    return 0;   
}
#endif

static inline void
set_opacity(FL_DFB_Window* window, int opacity){
  if (window) {
	window->setOpacity( opacity );
  }
}

static inline int
get_opacity(FL_DFB_Window* window){
  uchar ret_op = 255;
  if (window)
    ret_op = window->getOpacity();
  return ret_op;
}

static inline void
get_size(FL_DFB_Window* window, int* width, int* height){
  if (window)
    window->size(*width, *height);
}

static inline void
resize( FL_DFB_Window* window, int width, int height){
  if (window)
    window->resize(width, height);
}

static inline void
get_position(FL_DFB_Window* window,int* x, int* y){
  if (window)
    window->position(*x, *y);
}

static inline void
set_cursor_shape(IDirectFBDisplayLayer* layer, GC cursor, int hot_x, int hot_y){
  if (layer)
    DFBCHECK(layer->SetCursorShape(layer, cursor ? cursor->surface : 0, hot_x, hot_y));
}

/**********************************************************************/	

static inline void
wait_for_event_with_timeout(IDirectFBEventBuffer* buffer, unsigned sec, unsigned msec){
  if (buffer)
	DFBCHECK( buffer->WaitForEventWithTimeout( buffer, sec, msec ) );
}

static inline void
set_cursor_opacity(IDirectFBDisplayLayer* layer, int opacity){
  if (layer)
	DFBCHECK(layer->SetCursorOpacity(layer, opacity));
}

static inline void
get_cursor_position(IDirectFBDisplayLayer* layer, int* xx, int* yy){
  if (layer)
    DFBCHECK( layer->GetCursorPosition( layer, xx, yy ) );
}

GC
create_gc(const DFBSurfaceDescription* desc);

GC 
create_gc(FL_DFB_Window* window);

GC
create_gc(GC parent_gc, const DFBRectangle* rect);

GC
create_gc(int caps, int flags, int pixelformat, int width, int height);

GC
create_gc(IDirectFBDisplayLayer*);

#if 0
IDirectFBWindow* 
create_window(const DFBWindowDescription* desc);

IDirectFBWindow* 
get_window( DFBWindowID xid );

static inline void
grab_pointer(IDirectFBWindow* window){
  if (window)
    DFBCHECK( window->GrabPointer( window ) );
}

static inline void
ungrab_pointer(IDirectFBWindow* window){
  if (window)
    DFBCHECK(window->UngrabPointer(window));
}

static inline void
set_options(IDirectFBWindow* window, int opts){
  if (window)
    DFBCHECK(window->SetOptions(window, (DFBWindowOptions)opts));
}
#endif

static inline int
create_font(const char*	filename, const DFBFontDescription*	desc, IDirectFBFont** ret_interface){
  if (fl_dfb)
    fl_dfb->CreateFont(fl_dfb, filename, desc, ret_interface);
  return 0;   
}

static inline int
get_font_height( IDirectFBFont* font ){
    int height=0;
    if (font && ( DFB_OK == font->GetHeight(font, &height)))
        return height;
    return 0;   
}
    
static inline int
get_font_descender( IDirectFBFont* font ){
    int descender=0;
    if( font && ( DFB_OK == font->GetDescender( font, &descender ) ) )
        return -descender;
    return 0;   
}

static inline int
get_surface_description( IDirectFBImageProvider* image_provider, DFBSurfaceDescription* desc ){
  if (image_provider) 
	image_provider->GetSurfaceDescription( image_provider, desc );	
  return DFB_OK;
}

static inline IDirectFBImageProvider*
create_image_provider(const char *name){
  IDirectFBImageProvider *image_provider;
  if (fl_dfb && (DFB_OK == fl_dfb->CreateImageProvider(fl_dfb, name, &image_provider)) )
    return image_provider;
  return 0;
}
	
static inline void
render_to( IDirectFBImageProvider *image_provider, GC gc, DFBRectangle* rect ){
  if (image_provider)
	image_provider->RenderTo( image_provider, gc->surface, rect );
}

static inline void 
release( IDirectFBSurface* s ){
  if (s && fl_dfb)
    s->Release( s );
} 

#if 0
static inline void 
release( IDirectFBWindow *w ){
  if (w && fl_dfb)
    w->Release( w );
} 
#endif

static inline void 
release( IDirectFBInputDevice* d ){
  if (d && fl_dfb)
    d->Release( d );
} 

extern void print_gc_list();
extern void free_gc_list();
extern void fl_free_cursors();
extern void _fl_show_cursor(int);

#define fl_display    fl_display_layer
#define xdisplay      fl_display_layer
#define Fl_Region     Region
#define display_layer fl_display_layer

static inline int 
pixel_format(){
  extern u32 pixf_;
  return pixf_;
}

static inline bool
have_lut(){
  extern bool lut_;
  return lut_;
}

int fl_init_dfb(int argc, char **argv);

FL_EXPORT void fl_clip_region(Fl_Region);

#include "Fl_Window.H"
// this object contains all directfb-specific stuff about a window:
// Warning: this object is highly subject to change!
FL_EXPORT void fl_open_display();
FL_EXPORT void fl_open_display(Display*);
FL_EXPORT void fl_close_display();
FL_EXPORT bool fl_application_exited();
FL_EXPORT void fl_exit_application(int code = 0);

extern VisualInfo               visual_info;
extern FL_EXPORT dfbWindow      fl_message_window;
extern FL_EXPORT VisualInfo*    fl_visual;
extern FL_EXPORT Colormap 	    fl_colormap;
extern FL_EXPORT Surface        fl_gc;
extern FL_EXPORT Window         fl_window;
extern FL_EXPORT IDirectFBFont* fl_xfont;

extern FL_EXPORT ulong      fl_xpixel(Fl_Color i);
extern FL_EXPORT ulong      fl_xpixel(uchar r, uchar g, uchar b);
extern FL_EXPORT void       fl_clip_region(Fl_Region);
extern FL_EXPORT Fl_Region  fl_clip_region();

// feed events into fltk:
FL_EXPORT int fl_handle(const DFBWindowEvent&);

// you can use these in Fl::add_handler() to look at events:
//extern FL_EXPORT const DFBWindowEvent* fl_xevent;
extern FL_EXPORT ulong fl_event_time;
	
// flip a rectangle of screen 
FL_EXPORT void fl_flip( GC surf, int x, int y, int w, int h );
FL_EXPORT void fl_flip( int x, int y, int w, int h );
FL_EXPORT void fl_flip( GC surf, Region r );

/*******************************************************/
// off-screen pixmaps: create, destroy, draw into, copy to window:

typedef Pixmap Fl_Offscreen;

Fl_Offscreen fl_create_offscreen(int w,int h);
Fl_Offscreen fl_create_offscreen_argb(int w,int h);//with alpha mask

void fl_delete_offscreen(Fl_Offscreen);

// begin/end are functions that save the old state in global variables:

void fl_begin_offscreen(Fl_Offscreen);
void fl_end_offscreen();
void fl_copy_offscreen(int _x,int _y,int _w,int _h,Fl_Offscreen offscreen,int srcx,int srcy);
void fl_copy_offscreen_bk(int _x,int _y,int _w,int _h,Fl_Offscreen offscreen,int srcx,int srcy,ulong bkcolor );
//void fl_copy_offscreen_with_mask(int _x, int _y, int _w, int _h, void* offscreen, int srcx, int srcy, void* mask );
void fl_draw_abgr_to_argb(const uchar* array, int x, int y, int w, int h);
void fl_draw_argb_to_argb(const uchar* array, int x, int y, int w, int h);


#if 0 //FIXME SetDstColorKeyIndex is not working !
	extern int _dstcol_Blit( GC,GC,DFBRectangle*, int,int,int );
#   define __Blit( s ,i, r, x ,y, c ) _dstcol_Blit( s ,i, r, x ,y, c )
#else
#   define __Blit( s ,i, r, x ,y, c ) _Blit( s ,i, r, x ,y )
#endif

GC 
fl_load_image( const char *infname, int& w, int& h, GC* mask );

uchar*
fl_load_image( const char *infname, int& w, int& h, int& d );

int 
BitBlit( GC dest, int dest_x, int dest_y, int width, int height,GC src, int src_x,int src_y );

/*************************************************************************************/
/* Bitmap masks   */

typedef Fl_Offscreen Fl_Bitmask;

extern FL_EXPORT Fl_Bitmask fl_create_bitmask(int w, int h, const uchar *data);
extern FL_EXPORT Fl_Bitmask fl_create_alphamask(int w, int h, int d, int ld, const uchar *data);
extern FL_EXPORT void fl_delete_bitmask(Fl_Bitmask bm);

#define CHILD_ID 0x8000 

/*************************************************************************************/
// this object contains all X-specific stuff about a window:
// Warning: this object is highly subject to change!  It's definition
// is only here so that fl_xid can be declared inline:
class FL_EXPORT Fl_X {
public:
	int 		subWindow;
  	XID         xid;
  	dfbWindow  _xid;
  	Fl_Window*  w;
  	Fl_Region   region;
  	GC          gc;    //client's context
  	GC          win_gc;//window's context

  	char        wait_for_expose;
  	char        backbuffer_bad; // used for XDBE
  
  	static Fl_X* first;
  	Fl_X *next;
  	static Fl_X* i(const Fl_Window* wi) {return wi->i;}
/*------------------------------------------------------------------------*/  
  	void setwindow( Fl_Window* wi ) {w=wi; wi->i=this;}
  	void sendxjunk();
  	static Fl_X* make_xid(Fl_Window*);
  	static Fl_X* set_xid(Fl_Window*, dfbWindow);
   	char first_draw;//FIXME it's a flag to show the window after first drawing
	// kludges to get around protection:  	
  	void flush(); 
  	static void x(Fl_Window* wi, int X) {wi->x(X);}
  	static void y(Fl_Window* wi, int Y) {wi->y(Y);}
   	static void release();
   	
   	Fl_Window *header;
   	void draw_window_frame();
   	void enter_window(bool);
   	void set_gc(int W=-1,int H=-1);
   	DFBRectangle origin_size;
   	int minimize_;
   	int maximize(){return origin_size.w!=0;};
   	void set_maximize();
   	int minimize(){return minimize_;}
   	void delete_icon();
   	void set_minimize();
	void restore_size(int X=-1,int Y=-1,int W=-1,int H=-1);
	short minw,minh,maxw,maxh,dw,dh;
    Cursor cursors[2];
};

//compatibitily with FLTK2
#define CreatedWindow Fl_X

// convert xid <-> Fl_Window:
/*static inline WindowID fl_wnd(const Fl_Window* w){
    return Fl_X::i(w)->_xid;
}*/
static inline Window fl_xid(const Fl_Window* w){
    return Fl_X::i(w)->xid;
}

FL_EXPORT Fl_Window* fl_find( Window xid );
FL_EXPORT Fl_Window* fl_find( dfbWindow _xid );

Fl_X *get_parent_xid( Fl_Widget* w );

void delete_icon( Fl_Window* win );

extern FL_EXPORT char fl_override_redirect; // hack into Fl_X::make_xid()
extern FL_EXPORT int fl_background_pixel;  // hack into Fl_X::make_xid()

// Dummy function to register a function for opening files via the window manager...
static inline void fl_open_callback(void (*)(const char *)) {
}

extern FL_EXPORT int fl_parse_color(const char* p, uchar& r, uchar& g, uchar& b);
extern FL_EXPORT void fl_free_dfbfonts();

void fl_deinit_dfb(void);

#if USE_COLORMAP
void init_xcolormap();
#endif

//hack for dfb event.
extern FL_EXPORT bool fl_FLEventToDFBEvent(int e, DFBEvent& dfbevent);

//
// End of "$Id: DirectFB.H,v 1.3 2006-10-31 09:31:14 nikego Exp $".
//
