//
// "$Id xutils.h 4563 2006-03-06 23:03:34Z  $"
//
// DirectFB-specific code for the Fast Light Tool Kit (FLTK).
//
// Copyright 1998-2005 by Bill Spitzak and others.
//
// Written by Nikita Egorov (NikeGo@gmail.com) 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Library General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Library General Public License for more details.
//
// You should have received a copy of the GNU Library General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA.
//
// Please report all bugs and problems on the following page:
//
//     http://www.fltk.org/str.php
//

#ifndef _XUTILS_H_
#define _XUTILS_H_ 

#include <stdio.h> 
#include <malloc.h>

extern "C" {
#include <directfb.h>
#include <directfb_strings.h>
#include <directfb_util.h>
}


#include "Enumerations.H"

#if defined(USE_ALLOCA) && USE_ALLOCA
/* alloca() is available, so use it for better performance */
#define ALLOCA(size)    alloca(size)
#define FREEA(pmem)
#else
/* no alloca(), so use malloc()/free() instead */
#define ALLOCA(size)    Xmalloc(size)
#define FREEA(pmem) Xfree(pmem)
#endif

/* malloc stuff */
#if defined(MALLOC_0_RETURNS_NULL) && MALLOC_0_RETURNS_NULL
/* for machines that do not return a valid pointer for malloc(0)*/
# define Xmalloc(size) malloc(((size) == 0 ? 1 : (size)))
# define Xrealloc(ptr, size) realloc((ptr), ((size) == 0 ? 1 : (size)))
# define Xcalloc(nelem, elsize) calloc(((nelem) == 0 ? 1 : (nelem)), (elsize))
#else
# define Xmalloc(size) malloc((size))
# define Xrealloc(ptr, size) realloc((ptr), (size))
# define Xcalloc(nelem, elsize) calloc((nelem), (elsize))
#endif

#define Xfree(ptr) free((ptr))

typedef struct surface_context {
  IDirectFBSurface* surface;
  DFBUpdates        updates;
  surface_context*  clip_mask;
  DFBRegion         clip_region;
  struct {            
    u8 r,g,b,a;       // current color
  };
  struct {
    u32 dashmask;     // An actual bitmask of the dash values
    u32 dashcount;    // The number of bits defined in the dashmask
    u32 bit;          // current bit
 };
}*GC,*Surface;

class FL_DFB_Window;

typedef uintptr_t               XID;
typedef XID                     Window;
typedef XID                     Drawable;
typedef XID                     Pixmap;
typedef DFBPoint                XPoint;
typedef DFBPoint                POINT;

typedef DFBRectangle            RECT;
typedef DFBRectangle            XRectangle;
typedef DFBRegion*              Region;
typedef IDirectFBFont*          XFont;
typedef FL_DFB_Window*        dfbWindow;

typedef ulong                   Colormap;
typedef IDirectFBDisplayLayer   Display;

// try to convert xid into a pointer
#define ISGC(xid) ((xid)&~0xFFFF)?(GC)(xid):0

#define PIXMAP(surf) (Pixmap)(surf)

GC find_gc(XID xid);
    
typedef struct POINT_f {
    double x;
    double y;
};

/* outline and filled arc and pie types*/
#define ARC         0x0001  /* arc*/
#define OUTLINE     0x0002
#define ARCOUTLINE  0x0003  /* arc + outline*/
#define PIE         0x0004  /* pie (filled)*/
#define ELLIPSE     0x0008  /* ellipse outline*/
#define ELLIPSEFILL 0x0010  /* ellipse filled*/

/* set polygon fill routine*/
#define EDGEPOLYFILL    1   /* edge table, malloc, qsort*/
#define HAVEFLOAT       1

/* Fill mode  */
#define FILL_SOLID          0  
#define FILL_STIPPLE        1  
#define FILL_OPAQUE_STIPPLE 2  
#define FILL_TILE           3

#define MIN(a,b)      ((a) < (b) ? (a) : (b))
#define MAX(a,b)      ((a) > (b) ? (a) : (b))

#if 1
#  define MEMORY_TYPE DSCAPS_VIDEOONLY
#else
#  define MEMORY_TYPE DSCAPS_SYSTEMONLY
#endif

void release_gc(GC gc);
void release_window(dfbWindow win);

#define __tolower(a) (((a) >= 'A' && ((a) <= 'Z') ? ((a)+('a'-'A')):(a)))
#define __toupper(a) (((a) >= 'a' && ((a) <= 'z') ? ((a)-('a'-'A')):(a)))
 
/*
 * Where to find fonts...
 */
 
#define X11_FONT_DIR1 "/usr/fonts/truetype"
#define X11_FONT_DIR2 "/usr/X11R6/lib/X11/fonts/TrueType"

#define X11_RGBTXT "/usr/X11R6/lib/X11/rgb.txt"

/* Changeable options*/
#define USE_ALLOCA		1	/* set if system has alloca()*/
#define MALLOC_0_RETURNS_NULL	0	/* not yet needed*/

/* required settings*/
#define NeedFunctionPrototypes	1	/* ANSI C*/
#define XLIB_ILLEGAL_ACCESS	1	/* define real structures*/

/* deal with _Xconst differences in X11 header files*/
#ifndef XCONST
#define XCONST	_Xconst
#endif


#define None 0 

/*
 * cursor structure for DirectFB
 */
 
typedef struct _cursor {
	Pixmap shape;
	int hot_x;
	int hot_y;
    Pixmap source;
    Pixmap mask;
}*Cursor;

/* Display classes  used in opening the connection 
 * Note that the statically allocated ones are even numbered and the
 * dynamically changeable ones are odd numbered */

#define StaticGray		0
#define GrayScale		1
#define StaticColor		2
#define PseudoColor		3
#define TrueColor		4
#define DirectColor		5

#define LSBFirst		0
#define MSBFirst		1

/*
 * Visual structure; contains information about colormapping possible.
 */
typedef struct {
	void *ext_data;	/* hook for extension to hang data */
	int visualid;	/* visual id of this visual */
	int c_class;		/* C++ class of screen (monochrome, etc.) */
	unsigned long red_mask, green_mask, blue_mask;	/* mask values */
	int bits_per_rgb;	/* log base 2 of distinct color values */
	int map_entries;	/* color map entries */
} Visual;

/*
 * Information used by the visual utility routines to find desired visual
 * type from the many visuals a display may support.
 */
typedef struct {
  Visual *visual;
  int visualid;
  int screen;
  int depth;
  int c_class;					/* C++ */
  unsigned long red_mask;
  unsigned long green_mask;
  unsigned long blue_mask;
  int colormap_size;
  int bits_per_rgb;
  int byte_order;
///////////  
  int bits_per_pixel;
  int bytes_per_pixel;
} VisualInfo;

/* constants for bitmask operations*/
extern uchar BG_COLOR_IDX;
extern uchar MASK_COLOR_IDX;

extern uchar BG_COLOR_RED,BG_COLOR_GREEN,BG_COLOR_BLUE;
extern uchar MASK_COLOR_RED,MASK_COLOR_GREEN,MASK_COLOR_BLUE; 

extern u32 BG_COLOR_RGB32;
extern u32 MASK_COLOR_RGB32;

extern u16 BG_COLOR_RGB16;
extern u16 MASK_COLOR_RGB16;

#define Status int 
#define _Xconst const


// create compatible surface system memory only;
GC create_mem_gc( int w ,int h, int pf = -1 );
Pixmap create_pixmap( int w ,int h, int pf = -1 );

/*defines for unmodified (Xrm) Xlib routines...*/
/*#define LockDisplay(dpy)
#define UnlockDisplay(dpy)
#define _XLockMutex(lock)
#define _XUnlockMutex(lock)
#define _XCreateMutex(lock)
#define _XFreeMutex(lock)
*/
/* ImageFormat -- PutImage, GetImage */

#define XYBitmap		0	/* depth 1, XYFormat */
#define XYPixmap		1	/* depth == drawable depth */
#define ZPixmap			2	/* depth == drawable depth */

/*
 * Data structure for retrieving info about pixmap formats.
 */
typedef struct {
    int depth;
    int bits_per_pixel;
    int scanline_pad;
} XPixmapFormatValues;

/*
 * Data structure for "image" data, used by image manipulation routines.
 */
typedef struct _XImage {
    int width, height;		/* size of image */
    int xoffset;		/* number of pixels offset in X direction */
    int format;			/* XYBitmap, XYPixmap, ZPixmap */
    char *data;			/* pointer to image data */
    int byte_order;		/* data byte order, LSBFirst, MSBFirst */
    int bitmap_unit;		/* quant. of scanline 8, 16, 32 */
    int bitmap_bit_order;	/* LSBFirst, MSBFirst */
    int bitmap_pad;		/* 8, 16, 32 either XY or ZPixmap */
    int depth;			/* depth of image */
    int bytes_per_line;		/* accelarator to next line */
    int bits_per_pixel;		/* bits per pixel (ZPixmap) */
    unsigned long red_mask;	/* bits in z arrangment */
    unsigned long green_mask;
    unsigned long blue_mask;
  //  XPointer obdata;		/* hook for the object routines to hang on */
    struct funcs {		/* image manipulation routines */
#if  1 //NeedFunctionPrototypes
	struct _XImage *(*create_image)(
		Display* /* display */,
		Visual*		/* visual */,
		unsigned int	/* depth */,
		int		/* format */,
		int		/* offset */,
		char*		/* data */,
		unsigned int	/* width */,
		unsigned int	/* height */,
		int		/* bitmap_pad */,
		int		/* bytes_per_line */);
	int (*destroy_image)        (struct _XImage *);
	unsigned long (*get_pixel)  (struct _XImage *,unsigned int, unsigned int);
	int (*put_pixel)            (struct _XImage *,unsigned int, unsigned int, unsigned long);
	struct _XImage *(*sub_image)(struct _XImage *,unsigned int, unsigned int, unsigned int, unsigned int);
	int (*add_pixel)            (struct _XImage *, long);
#else
	struct _XImage *(*create_image)();
	int (*destroy_image)();
	unsigned long (*get_pixel)();
	int (*put_pixel)();
	struct _XImage *(*sub_image)();
	int (*add_pixel)();
#endif
	} f;
} XImage;

/*
 *  stubs..
 */ 

#define  	XKeysymToString( key ) 0
#define		XBell( a, b ) 

#define XMapWindow( d, id) /*ShowWindow(b, SW_RESTORE)*/
#define XMapRaised( d, id)
#define XUnmapWindow( d, id) /*ShowWindow(b, SW_HIDE)*/

void XFlush( Display* display);

static inline void 
redraw() { //FIXME
}

#define XGetDefault(fl_display, key1, key2) 0
#define RootWindow( d, s ) fl_window
#define DefaultRootWindow( d ) fl_window

static inline int
XQueryKeymap( Display* d, char s[32]){
    printf("XQueryKeymap called\n");
    return 0;
} 

int 
XKeysymToKeycode( Display* d, int ks );


#define XNegative 0
#define YNegative 0
#define XValue 0
#define YValue 0

#define AllocNone 0

#define AllPlanes 		((unsigned long)~0L)

extern int gr_fillmode;

/*
 * Data structure used by color operations
 */
typedef struct {
	long pixel;
	unsigned short red, green, blue;
	unsigned short alpha;
	char flags;  /* do_red, do_green, do_blue */
	char pad;
} XColor;

/* Used internally for the colormap */

#define GR_PIXELVAL int
#define GR_COLOR int

typedef struct  {
	GR_PIXELVAL	value;
	int		ref;
	unsigned short r, g, b;
} nxColorval;

typedef struct _nxColormap {
	int		id;
	int		color_alloc;
	int		cur_color;
	int 	saved_color;
	DFBColor	colors[256];	
	nxColorval *		colorval;
	struct _nxColormap *	next;
} nxColormap;

extern Pixmap clip_mask;
/*
 * 
 *  Region and rectange functions
 * 
 */
int 
XCopyArea(Display *display,Drawable src,Drawable dest,GC gc,
    int src_x,int src_y,unsigned width,unsigned height,int dest_x,int dest_y);
 
Region 
XRectangleRegion(int x, int y, int w, int h);

Region 
XCreateRegion();

void 
XDestroyRegion(Region r);

int 
XRectInRegion(Region r,int x,int y,int w,int h);

void
XClipBox(Region r, XRectangle* rect );

void
XSetRegion(Display* display, GC gc, Region r, int set_clip=1);

void
XSetClipMask(Display* display, GC gc, Pixmap pixmap);

int
XIntersectRegion(Region r1, Region r2, Region newReg);

int
XSetClipOrigin(Display* display, GC gc, int clip_x_origin, int clip_y_origin);

int
XUnionRectWithRegion(XRectangle* rect, Region source, Region dest);

int
XUnionRectWithRegion( int x,int y,int w,int h, Region source, Region dest);

int
XParseGeometry(const char *parsestring, int *x, int *y, unsigned int *w, unsigned int *h);

Status
XParseColor(Display *display, Colormap colormap, const char *spec, XColor *exact_def_return);

void
flip_updates( GC );

void
remove_updates( GC );

void
update_region( GC surface, DFBRegion* r );

void
update_region(GC surface,int x1,int y1,int x2,int y2);

#define update_rect(s,x,y,w,h)\
    update_region((s),DFB_REGION_VALS_FROM_RECTANGLE_VALS((x),(y),(w),(h)));

int
UnionReg( Region r1/*in*/, Region r2/*in/out*/ );

void 
XDrawLine(Display* display,Drawable d,GC gc,int x1,int y1,int x2,int y2);

void 
XDrawLines(Display* display,Drawable d,GC gc,Region points,int npoints,int mode);

void 
XDrawRectangle(Display* display,Drawable d,GC gc,int x,int y,int width,int height);

void 
XFillRectangle(Display* display,Drawable d,GC gc,int x,int y,int width,int height);

#define draw_line_i(x1,y1,x2,y2)\
	XDrawLine(fl_display,fl_window,fl_gc, (int)(x1+.5),(int)(y1+.5),(int)(x2+.5),(int)(y2+.5))

#define draw_line_r(x1,y1,x2,y2)\
	draw_smooth_line(fl_gc,x1,y1,x2,y2)
    
void
FillPoly(Drawable d,GC gc,int count, POINT_f *pointtable );

void
SmoothPoly(GC gc, int count, POINT_f *pointtable);

void
Arc(GC gc, short x0, short y0, short rx, short ry,    short ax, short ay, short bx, short by, int type);

void
ArcAngle(GC gc, int x,int y,int w,int h,double a1,double a2, int type );

void 
draw_smooth_line(GC gc, double x1, double y1, double x2, double y2);

void
fl_filter_offscreen(GC offscreen,int x,int y,int w,int h,int );

void
drawrow(GC gc,int x1,int x2,int y);

int
dash(GC ctx);

void
set_dash_style(GC ctx, u32 mask, u32 count);

void
reset_dash(GC ctx);

void
SetDashStyle( unsigned long mask );

int
SetClipRect ( int x, int y, int w, int h );
int
SetClipRect ( DFBRegion* r );

/* _colormap.cxx*/
nxColormap
*_nxFindColormap(Colormap id);

Colormap
_nxDefaultColormap(Display* display);

Colormap
XCreateColormap(Display * display, Window w, Visual * visual, int alloc);

inline Colormap
XCreateColormap(Display *display ){
	return XCreateColormap( display , 0, 0, 0 );
}

GR_COLOR
GrGetColorByName(char *colorname, int *retr, int *retg, int *retb);

int
get_rgb_color( int xcolor, uchar& r, uchar& g, uchar& b );

int
get_index_color(uchar r,uchar g,uchar b);

int
XAllocColor( Display* display, Colormap colormap, XColor* in_out);

int
XQueryColor( Display* display, Colormap colormap, XColor * def);

int
XQueryColors( Display* display, Colormap colormap, XColor * def_in,int ncolors);

int
XFreeColors( Display* display, Colormap colormap, unsigned long pixels[],int npixels, unsigned long planes);

int
XFlushColormap( Display* display, Colormap colormap );

void _nxPixel2RGB(Display* display, unsigned long color,
	   unsigned short *red, unsigned short *green, unsigned short *blue);

/* _listfonts.cxx */
extern char **_nxfontlist;

extern int _nxfontcount;

FILE * 
_nxLoadFontDir(char *str);

void 
_nxSetDefaultFontDir(void);

void 
_nxSetFontDir(char **directories, int ndirs);

char** 
_nxGetFontDir(int *count);

void 
_nxFreeFontDir(char **list);

/* _loadfont.cxx*/
char *
_nxFindX11Font(const char *in_font);

/*  */
char **
XListFonts( Display* display, const char *pattern, int maxnames, int *actual_count_return);

XFont 
XLoadFont( Display* display, const char *fname );

void 
XDestroyWindow( Display* display , Window window );

static inline int
XFreeFont( Display* d, XFont font){
	if (font)
    font->Release(font);
  return 0;    
}

int 
XFreeFontNames(char **list);

int 
XIconifyWindow(Display* d, Window w, void* v);

/*
 *  image functions 
 */
int
XPutImage(Display* display, Drawable d, GC gc, XImage * image,
	int src_x, int src_y, int dest_x, int dest_y, unsigned int width,
	unsigned int height);

XImage*	
XCreateImage(Display* display, Visual* visual, unsigned int depth,
	int format, int offset, char *data, unsigned int width,
	unsigned int height, int bitmap_pad, int bytes_per_line);
	
XImage*
XGetImage(Display* display, Drawable d, int x, int y,
	unsigned int width, unsigned int height,
	unsigned long plane_mask, int format);

int
XDestroyImage(XImage *image);

Pixmap
XCreateBitmapFromData(Display* display, Drawable d, const char *data,
	unsigned int width, unsigned int height);

/* 
 * cursors 
 */

void 
XFreeCursor( Display* display, Cursor xc );

inline void 
XFreePixmap( Display* display, Pixmap p )
{
  if (p & 0xFFFF)
    release_gc( (GC)p );
} 

int
XSetForeground( Display* display, GC gc, int color_index );
	
void
XDefineCursor( Display* display, Window win, Cursor c );

Cursor
XCreatePixmapCursor( Display* display, Pixmap p, Pixmap m, XColor *col1, XColor *col2, unsigned hot_x, unsigned hot_y );

Cursor
XCreateFontCursor( Display *display, unsigned int shape );

int 
XRecolorCursor( Display* display, Cursor cursor, XColor *fg_col, XColor *bg_col );

/********************************************************************************/

/*static inline int
__sleep(int mssec){
    struct timespec rqtp,rmtp;
    rqtp.tv_sec  = mssec / 1000;
    rqtp.tv_nsec = mssec * 1000000;
    nanosleep(&rqtp, &rmtp);
    return rmtp.tv_sec * 1000 + rmtp.tv_nsec / 1000000;
}*/

static const 
DirectFBPixelFormatNames(format_names);

static inline const char *
pixelformat_name( int format ){
     int i;
     for (i=0; format_names[i].format != DSPF_UNKNOWN; i++) {
          if (format_names[i].format == (DFBSurfacePixelFormat)format)
               return format_names[i].name;
     }
     return "INVALID FORMAT";
}

void dump_surf_data( GC surf, int x, int y, int w, int h );

/***************************************************************** 
 * EVENT DEFINITIONS 
 *****************************************************************/

/* Input Event Masks. Used as event-mask window attribute and as arguments
   to Grab requests.  Not to be confused with event names.  */

#define NoEventMask         0L
#define KeyPressMask            (1L<<0)  
#define KeyReleaseMask          (1L<<1)  
#define ButtonPressMask         (1L<<2)  
#define ButtonReleaseMask       (1L<<3)  
#define EnterWindowMask         (1L<<4)  
#define LeaveWindowMask         (1L<<5)  
#define PointerMotionMask       (1L<<6)  
#define PointerMotionHintMask       (1L<<7)  
#define Button1MotionMask       (1L<<8)  
#define Button2MotionMask       (1L<<9)  
#define Button3MotionMask       (1L<<10) 
#define Button4MotionMask       (1L<<11) 
#define Button5MotionMask       (1L<<12) 
#define ButtonMotionMask        (1L<<13) 
#define KeymapStateMask         (1L<<14)
#define ExposureMask            (1L<<15) 
#define VisibilityChangeMask        (1L<<16) 
#define StructureNotifyMask     (1L<<17) 
#define ResizeRedirectMask      (1L<<18) 
#define SubstructureNotifyMask      (1L<<19) 
#define SubstructureRedirectMask    (1L<<20) 
#define FocusChangeMask         (1L<<21) 
#define PropertyChangeMask      (1L<<22) 
#define ColormapChangeMask      (1L<<23) 
#define OwnerGrabButtonMask     (1L<<24) 

/* Event names.  Used in "type" field in XEvent structures.  Not to be
confused with event masks above.  They start from 2 because 0 and 1
are reserved in the protocol for errors and replies. */

#define KeyPress        2
#define KeyRelease      3
#define ButtonPress     4
#define ButtonRelease       5
#define MotionNotify        6
#define EnterNotify     7
#define LeaveNotify     8
#define FocusIn         9
#define FocusOut        10
#define KeymapNotify        11
#define Expose          12
#define GraphicsExpose      13
#define NoExpose        14
#define VisibilityNotify    15
#define CreateNotify        16
#define DestroyNotify       17
#define UnmapNotify     18
#define MapNotify       19
#define MapRequest      20
#define ReparentNotify      21
#define ConfigureNotify     22
#define ConfigureRequest    23
#define GravityNotify       24
#define ResizeRequest       25
#define CirculateNotify     26
#define CirculateRequest    27
#define PropertyNotify      28
#define SelectionClear      29
#define SelectionRequest    30
#define SelectionNotify     31
#define ColormapNotify      32
#define ClientMessage       33
#define MappingNotify       34
#define LASTEvent       35  /* must be bigger than any event # */

/*
 * Definitions of specific events.
 */
#if 0 
typedef struct {
    int type;       /* of event */
    unsigned long serial;   /* # of last request processed by server */
    bool send_event;    /* true if this came from a SendEvent request */
    Display *display;   /* Display the event was read from */
    Window window;          /* "event" window it is reported relative to */
    Window root;            /* root window that the event occurred on */
    Window subwindow;   /* child window */
    Time time;      /* milliseconds */
    int x, y;       /* pointer x, y coordinates in event window */
    int x_root, y_root; /* coordinates relative to root */
    unsigned int state; /* key or button mask */
    unsigned int keycode;   /* detail */
    Bool same_screen;   /* same screen flag */
} XKeyEvent;

typedef XKeyEvent XKeyPressedEvent;
typedef XKeyEvent XKeyReleasedEvent;

typedef struct {
    int type;       /* of event */
    unsigned long serial;   /* # of last request processed by server */
    bool send_event;    /* true if this came from a SendEvent request */
    Display *display;   /* Display the event was read from */
    Window window;          /* "event" window it is reported relative to */
    Window root;            /* root window that the event occurred on */
    Window subwindow;   /* child window */
    Time time;      /* milliseconds */
    int x, y;       /* pointer x, y coordinates in event window */
    int x_root, y_root; /* coordinates relative to root */
    unsigned int state; /* key or button mask */
    unsigned int button;    /* detail */
    Bool same_screen;   /* same screen flag */
} XButtonEvent;
typedef XButtonEvent XButtonPressedEvent;
typedef XButtonEvent XButtonReleasedEvent;

typedef struct {
    int type;       /* of event */
    unsigned long serial;   /* # of last request processed by server */
    bool send_event;    /* true if this came from a SendEvent request */
    Display *display;   /* Display the event was read from */
    Window window;          /* "event" window reported relative to */
    Window root;            /* root window that the event occurred on */
    Window subwindow;   /* child window */
    Time time;      /* milliseconds */
    int x, y;       /* pointer x, y coordinates in event window */
    int x_root, y_root; /* coordinates relative to root */
    unsigned int state; /* key or button mask */
    char is_hint;       /* detail */
    Bool same_screen;   /* same screen flag */
} XMotionEvent;
typedef XMotionEvent XPointerMovedEvent;

#endif //if 0

typedef struct {
    int type;
    unsigned long serial;   /* # of last request processed by server */
    bool send_event;    /* true if this came from a SendEvent request */
    Display *display;   /* Display the event was read from */
    Window window;
    int x, y;
    int width, height;
    int count;      /* if non-zero, at least this many more */
} XExposeEvent;

typedef struct {
    int type;
    unsigned long serial;   /* # of last request processed by server */
    bool send_event;    /* true if this came from a SendEvent request */
    Display *display;   /* Display the event was read from */
    Drawable drawable;
    int x, y;
    int width, height;
    int count;      /* if non-zero, at least this many more */
    int major_code;     /* core is CopyArea or CopyPlane */
    int minor_code;     /* not defined in the core */
} XGraphicsExposeEvent;

typedef struct {
    int type;
    unsigned long serial;   /* # of last request processed by server */
    bool send_event;    /* true if this came from a SendEvent request */
    Display *display;   /* Display the event was read from */
    Drawable drawable;
    int major_code;     /* core is CopyArea or CopyPlane */
    int minor_code;     /* not defined in the core */
} XNoExposeEvent;

typedef struct {
    int type;
    unsigned long serial;   /* # of last request processed by server */
    bool send_event;    /* true if this came from a SendEvent request */
    Display *display;/* Display the event was read from */
    Window window;  /* window on which event was requested in event mask */
} XAnyEvent;

/*
 * this union is defined so Xlib can always use the same sized
 * event structure internally, to avoid memory fragmentation.
 */
typedef union _XEvent {
        int type;       /* must not be changed; first element */
    XAnyEvent xany;
//    XKeyEvent xkey;
//    XButtonEvent xbutton;
//    XMotionEvent xmotion;
    XExposeEvent xexpose;
    XGraphicsExposeEvent xgraphicsexpose;
    XNoExposeEvent xnoexpose;
    long pad[24];
} XEvent;

int
XWindowEvent(Display* display, Window w, long event_mask, XEvent * ev);

#define PIXEL_RGB332(r,g,b)    ( (((r)&0xE0)     ) | \
                                 (((g)&0xE0) >> 3) | \
                                 (((b)&0xC0) >> 6) )
                                 
#define PIXEL_RGB16(r,g,b)     ( (((r)&0xF8) << 8) | \
                                 (((g)&0xFC) << 3) | \
                                 (((b)&0xF8) >> 3) )

#define RGB16_RED(p)           (((p)&0xF800 ) >> 8 )

#define RGB16_GREEN(p)         (((p)&0x07E0 ) >> 3 )

#define RGB16_BLUE(p)          (((p)&0x001F ) << 3 )

#define PIXEL_RGB32(r,g,b)     ( ((r) << 16) | \
                                 ((g) <<  8) | \
                                  (b) )

#define PIXEL_ARGB(a,r,g,b)    ( ((a) << 24) | \
                                 ((r) << 16) | \
                                 ((g) << 8)  | \
                                  (b) )
                                  
#define PIXEL_ARGB4444(a,r,g,b)( (((a)&0xF0) << 8) | \
                                 (((r)&0xF0) << 4) | \
                                 (((g)&0xF0)     ) | \
                                 (((b)&0xF0) >> 4) )
                                  

#define ARGB_ALPHA(pixelval)	((pixelval & 0xFF000000) >> 24)

#define ARGB_RED(pixelval)  	((pixelval & 0x00FF0000) >> 16)

#define ARGB_GREEN(pixelval)	((pixelval & 0x0000FF00) >> 8)

#define ARGB_BLUE(pixelval) 	 (pixelval & 0x000000FF)

#define RGB32_RED(pixelval)	    ((pixelval & 0x00FF0000) >> 16)

#define RGB32_GREEN(pixelval)   ((pixelval & 0x0000FF00) >> 8)

#define RGB32_BLUE(pixelval)	 (pixelval & 0x000000FF)

/* more fast macroses - operate with addresses */

#define LOAD_U16_VO( addr )        *((u16*)(addr))

#define STORE_U16_VO( addr, val )  *((u16*)(addr)) = (val)

#define LOAD_U32_VO( addr )        *((u32*)(addr))

#define STORE_U32_VO( addr, val)   *((u32*)(addr)) = (val)

#ifdef WORDS_BIGENDIAN
#  if WORDS_BIGENDIAN == 0
#    undef WORDS_BIGENDIAN
#  endif
#endif

#ifdef WORDS_BIGENDIAN

/* architecture dependent macroses. objects to change */

#if (DIRECTFB_MAJOR_VERSION != 0) //BDH sharp
#include <linux/config.h>
#endif

#ifdef REVERSE_BYTES /* e.g. 32 bits BGRA pixel format */

#undef LOAD_U16_VO
#define LOAD_U16_VO( addr )        load_le16( (u16*)(addr) )

#undef STORE_U16_VO
#define STORE_U16_VO( addr, val )  store_le16( (u16*)(addr), (val) )

#undef LOAD_U32_VO
#define LOAD_U32_VO( addr )        load_le32( (u32*)(addr) )

#undef STORE_U32_VO
#define STORE_U32_VO( addr, val)   store_le32( (u32*)addr, (val) )

#if defined( CONFIG_PPC ) || defined( CONFIG_POWERPC ) /* PowerPC architecture */

/*
 * 16 and 32 bit, big and little endian I/O operations.
 */
 
inline static 
u16 load_le16(const u16 *addr){
    u16 ret;
    __asm__ __volatile__("lhbrx %0,0,%1;\n"
                  : "=r" (ret) :
                  "r" (addr), "m" (*addr));
    return ret;
}

inline static 
void store_le16( u16 *addr, u16 val){
    __asm__ __volatile__("sthbrx %1,0,%2;\n" : "=m" (*addr) :
                  "r" (val), "r" (addr));
}

inline static 
u32 load_le32(const u32 *addr){
    u32 ret;

    __asm__ __volatile__("lwbrx %0,0,%1;\n"
                 : "=r" (ret) :
                 "r" (addr), "m" (*addr));
    return ret;
}

inline static 
void store_le32( u32 *addr, u32 val){
    __asm__ __volatile__("stwbrx %1,0,%2;\n" : "=m" (*addr) :
                 "r" (val), "r" (addr));
}

#else /* not PowerPC architecture */
/* */
static inline 
u16 load_le16( u16* p ){
    return ( *p << 8) | ( *p >> 8 );
}

static inline 
u32 load_le32( u32* p ){
    u32 val = (*p & 0x00FF00FF) << 8 | (*p & 0xFF00FF00) >> 8;
    return (val & 0x0000FFFF) << 16| (val & 0xFFFF0000) >> 16; 
}
  
static inline 
void store_u16( u16* p, u16 val ){
    *p = ( val << 8) | ( val >> 8 );
}

static inline 
void store_u32( u32* p, u32 val ){
        val = (val & 0x00FF00FF) << 8 | (val & 0xFF00FF00) >> 8;
        *p  = (val & 0x0000FFFF) << 16| (val & 0xFFFF0000) >> 16; 
}
        
#endif /* of PowerPC architecture */

inline static
u16 reverse_u16( u16 val ){
    return load_le16( &val );
}

inline static
u32 reverse_u32( u32 val ){
    return load_le32( &val );
}

#endif //REVERSE COLORS
#endif // WORDS_BIGENDIAN


#endif /* _XUTILS_H_*/
