//
// "$Id: _draw.cxx,v 1.1 2006-11-02 08:53:35 nikego Exp $"
//
// DirectFB-specific code for the Fast Light Tool Kit (FLTK).
//
// Copyright 1998-2005 by Bill Spitzak and others.
// Written by Nikita Egorov (nikego@gmail.com)
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Library General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Library General Public License for more details.
//
// You should have received a copy of the GNU Library General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA.
//
// Please report all bugs and problems on the following page:
//
//     http://www.fltk.org/str.php
//
// This file contains drawing code for fltk which is always linked
// in.  Search other files for "DIRECTFB" or filenames ending in _directfb.cxx
// for other system-specific code.
// 

/*********************************************************************/
/* those functions are defined in DirectFB's folder src/gfx/clip.h   */
/* they are internal functions...                                    */

typedef enum {
     DFEF_NONE      = 0x00000000,

     DFEF_LEFT      = 0x00000001,
     DFEF_RIGHT     = 0x00000002,
     DFEF_TOP       = 0x00000004,
     DFEF_BOTTOM    = 0x00000008,

     DFEF_ALL       = 0x0000000F
} DFBEdgeFlags;

extern "C" DFBEdgeFlags dfb_clip_edges(const DFBRegion* clip, DFBRectangle* rect);

extern "C" DFBBoolean   dfb_clip_line(const DFBRegion* clip, DFBRegion* line);

/* DirectFB's folder src/gfx/convert.h   */

#if (DIRECTFB_MAJOR_VERSION < 1 ||  DIRECTFB_MINOR_VERSION <= 0)
extern "C" u32 dfb_color_to_pixel( DFBSurfacePixelFormat format, u8 r, u8 g, u8 b );
#else
extern "C" u32 dfb_pixel_from_color( DFBSurfacePixelFormat format, DFBColor *color);

static u32 dfb_color_to_pixel( DFBSurfacePixelFormat format, u8 r, u8 g, u8 b)
{
    DFBColor color = {255, r, g, b};
    return dfb_pixel_from_color(format, &color);
}
#endif

/*********************************************************************/
/*
 * Simple class for draw some stuff directly in memory  
 */

class direct_draw{
    GC mgc;
    uchar* mbuf,*mbm;
    int mpixf;
    int mpitch,mcolor,mw,mh,mbmw,mbpp;
public: 
    direct_draw(GC gc);
    direct_draw(GC gc,uchar* bm);
    virtual ~direct_draw();
    void begin_draw(GC gc); 
    void end_draw();
    int getpixel(int,int) const;
    void setpixel(int,int);
    void setpixel(int x,int y,int color);
    void setpixel(int x,int y,double br);
    void setcolor(u8,u8,u8,u8);
    void setrgb(int,int,int);
    void setrow(int x,int y,int w);
    void fillrow( int& left, int& right,int y,ulong border_color);
    void fillrow(int& left, int& right,int y);
    void setcol(int x,int y,int h);
    void setbit(int x,int y);
    int getbit(int x,int y);
private:
    void _setbit(int x,int y);
    int _getbit(int x,int y);
};

//**************************************************************************
void draw_line(GC ctx, DFBRegion *line);

int
dash(GC ctx)
{
  if(!ctx)
    return 0;
  int on = (ctx->dashmask & (1 << ctx->bit)) ? 1 : 0;
  ctx->bit = (ctx->bit + 1) % ctx->dashcount;
  return on;
}

void
set_dash_style(GC ctx, u32 mask, u32 count)
{
  if (ctx)
  {
    ctx->dashmask  = mask;
    ctx->dashcount = count;
  }
}

void
reset_dash(GC ctx)
{
  if (ctx)
    ctx->bit = 0;
}

/******************************************************************
 * overrided drawing functions
 */
void
XDrawLine(Display* display, Drawable d, GC gc, int x1, int y1, int x2, int y2)
{
  GC ctx = ISGC(d);
  if (!ctx)
    ctx = gc;
    
  if (ctx->dashmask != 0xFFFFFFFF && ctx->dashmask != 0){
    DFBRegion line = {x1:x1, y1:y1, x2:x2, y2:y2};
    draw_line(ctx, &line);
  }else{
    DFBCHECK(ctx->surface->DrawLine(ctx->surface, x1, y1, x2, y2));
  }
  update_region(ctx, x1, y1, x2, y2);
}

void 
XDrawLines(Display* display, Drawable d, GC gc, Region points, int npoints, int mode)
{
  GC ctx = ISGC(d);
  if (!ctx)
    ctx = gc;
  
  int i;
  DFBRegion* reg = points;
  if (ctx->dashmask != 0xFFFFFFFF && ctx->dashmask != 0){
    for (i = 0; i < npoints; i++){
      draw_line(ctx, reg);
      reg++;
    }
  }else{
    DFBCHECK(ctx->surface->DrawLines(ctx->surface, points, npoints));
  }
  reg = points;
  for (i = 0; i < npoints; i++)
    update_region(ctx, reg++);
}
    
void 
XDrawRectangle(Display* display, Drawable d, GC gc, int x, int y, int width, int height)
{
  GC ctx = ISGC(d);
  if (!ctx)
    ctx = gc;
  
  if (ctx->dashmask != 0xFFFFFFFF && ctx->dashmask != 0) {
    direct_draw dd(ctx);
    DFBRegion   clip;
    reset_dash(ctx);
    DFBRectangle rect = {x, y, width, height};
    get_clip(ctx, &clip);
    if (clip.x2 > clip.x1)
      clip.x2--;//FIXME
    if (clip.y2 > clip.y1)
      clip.y2--;//FIXME

    DFBEdgeFlags edge_flags = dfb_clip_edges(&clip, &rect);
    if (edge_flags & DFEF_TOP)
      dd.setrow(rect.x, rect.y, rect.w);

    if (edge_flags & DFEF_RIGHT)
      dd.setcol(rect.x + rect.w - 1, rect.y, rect.h);

    if (edge_flags & DFEF_BOTTOM)
      dd.setrow(rect.x + rect.w - 1, rect.y + rect.h, -rect.w);

    if (edge_flags & DFEF_LEFT)
      dd.setcol(rect.x, rect.y + rect.h - 1, -(rect.h));

    dd.end_draw();
    update_region(ctx, DFB_REGION_VALS_FROM_RECTANGLE(&rect));
  }else{
    DFBCHECK(ctx->surface->DrawRectangle(ctx->surface, x, y, ++width, ++height));
    update_rect(ctx, x, y, width, height);
  }
}

void
XFillRectangle(Display* display, Drawable d, GC gc, int x, int y, int width, int height)
{
  GC ctx = ISGC(d);
  if (!ctx)
    ctx = gc;
  
  DFBCHECK(ctx->surface->FillRectangle(ctx->surface, x, y, width, height));
  update_rect(ctx, x, y, width, height);       
}

//************************************************************************

// fl_encoding variable defined in fl_font_directfb.cxx
extern const char* fl_encoding;

// cyrillic charsets conversions to unicode. The tables were taken from "unicode.org".
static uint16_t cp1251_to_uni[256] ={\
0x0000,0x0001,0x0002,0x0003,0x0004,0x0005,0x0006,0x0007,0x0008,0x0009,0x000A,0x000B,0x000C,0x000D,0x000E,0x000F,\
0x0010,0x0011,0x0012,0x0013,0x0014,0x0015,0x0016,0x0017,0x0018,0x0019,0x001A,0x001B,0x001C,0x001D,0x001E,0x001F,\
0x0020,0x0021,0x0022,0x0023,0x0024,0x0025,0x0026,0x0027,0x0028,0x0029,0x002A,0x002B,0x002C,0x002D,0x002E,0x002F,\
0x0030,0x0031,0x0032,0x0033,0x0034,0x0035,0x0036,0x0037,0x0038,0x0039,0x003A,0x003B,0x003C,0x003D,0x003E,0x003F,\
0x0040,0x0041,0x0042,0x0043,0x0044,0x0045,0x0046,0x0047,0x0048,0x0049,0x004A,0x004B,0x004C,0x004D,0x004E,0x004F,\
0x0050,0x0051,0x0052,0x0053,0x0054,0x0055,0x0056,0x0057,0x0058,0x0059,0x005A,0x005B,0x005C,0x005D,0x005E,0x005F,\
0x0060,0x0061,0x0062,0x0063,0x0064,0x0065,0x0066,0x0067,0x0068,0x0069,0x006A,0x006B,0x006C,0x006D,0x006E,0x006F,\
0x0070,0x0071,0x0072,0x0073,0x0074,0x0075,0x0076,0x0077,0x0078,0x0079,0x007A,0x007B,0x007C,0x007D,0x007E,0x007F,\
0x0402,0x0403,0x201A,0x0453,0x201E,0x2026,0x2020,0x2021,0x20AC,0x2030,0x0409,0x2039,0x040A,0x040C,0x040B,0x040F,\
0x0452,0x2018,0x2019,0x201C,0x201D,0x2022,0x2013,0x2014,0x0000,0x2122,0x0459,0x203A,0x045A,0x045C,0x045B,0x045F,\
0x00A0,0x040E,0x045E,0x0408,0x00A4,0x0490,0x00A6,0x00A7,0x0401,0x00A9,0x0404,0x00AB,0x00AC,0x00AD,0x00AE,0x0407,\
0x00B0,0x00B1,0x0406,0x0456,0x0491,0x00B5,0x00B6,0x00B7,0x0451,0x2116,0x0454,0x00BB,0x0458,0x0405,0x0455,0x0457,\
0x0410,0x0411,0x0412,0x0413,0x0414,0x0415,0x0416,0x0417,0x0418,0x0419,0x041A,0x041B,0x041C,0x041D,0x041E,0x041F,\
0x0420,0x0421,0x0422,0x0423,0x0424,0x0425,0x0426,0x0427,0x0428,0x0429,0x042A,0x042B,0x042C,0x042D,0x042E,0x042F,\
0x0430,0x0431,0x0432,0x0433,0x0434,0x0435,0x0436,0x0437,0x0438,0x0439,0x043A,0x043B,0x043C,0x043D,0x043E,0x043F,\
0x0440,0x0441,0x0442,0x0443,0x0444,0x0445,0x0446,0x0447,0x0448,0x0449,0x044A,0x044B,0x044C,0x044D,0x044E,0x044F};

static uint16_t koi8r_to_uni[256] = {\
0x0000,0x0001,0x0002,0x0003,0x0004,0x0005,0x0006,0x0007,0x0008,0x0009,0x000A,0x000B,0x000C,0x000D,0x000E,0x000F,\
0x0010,0x0011,0x0012,0x0013,0x0014,0x0015,0x0016,0x0017,0x0018,0x0019,0x001A,0x001B,0x001C,0x001D,0x001E,0x001F,\
0x0020,0x0021,0x0022,0x0023,0x0024,0x0025,0x0026,0x0027,0x0028,0x0029,0x002A,0x002B,0x002C,0x002D,0x002E,0x002F,\
0x0030,0x0031,0x0032,0x0033,0x0034,0x0035,0x0036,0x0037,0x0038,0x0039,0x003A,0x003B,0x003C,0x003D,0x003E,0x003F,\
0x0040,0x0041,0x0042,0x0043,0x0044,0x0045,0x0046,0x0047,0x0048,0x0049,0x004A,0x004B,0x004C,0x004D,0x004E,0x004F,\
0x0050,0x0051,0x0052,0x0053,0x0054,0x0055,0x0056,0x0057,0x0058,0x0059,0x005A,0x005B,0x005C,0x005D,0x005E,0x005F,\
0x0060,0x0061,0x0062,0x0063,0x0064,0x0065,0x0066,0x0067,0x0068,0x0069,0x006A,0x006B,0x006C,0x006D,0x006E,0x006F,\
0x0070,0x0071,0x0072,0x0073,0x0074,0x0075,0x0076,0x0077,0x0078,0x0079,0x007A,0x007B,0x007C,0x007D,0x007E,0x007F,\
0x2500,0x2502,0x250C,0x2510,0x2514,0x2518,0x251C,0x2524,0x252C,0x2534,0x253C,0x2580,0x2584,0x2588,0x258C,0x2590,\
0x2591,0x2592,0x2593,0x2320,0x25A0,0x2219,0x221A,0x2248,0x2264,0x2265,0x00A0,0x2321,0x00B0,0x00B2,0x00B7,0x00F7,\
0x2550,0x2551,0x2552,0x0451,0x2553,0x2554,0x2555,0x2556,0x2557,0x2558,0x2559,0x255A,0x255B,0x255C,0x255D,0x255E,\
0x255F,0x2560,0x2561,0x0401,0x2562,0x2563,0x2564,0x2565,0x2566,0x2567,0x2568,0x2569,0x256A,0x256B,0x256C,0x00A9,\
0x044E,0x0430,0x0431,0x0446,0x0434,0x0435,0x0444,0x0433,0x0445,0x0438,0x0439,0x043A,0x043B,0x043C,0x043D,0x043E,\
0x043F,0x044F,0x0440,0x0441,0x0442,0x0443,0x0436,0x0432,0x044C,0x044B,0x0437,0x0448,0x044D,0x0449,0x0447,0x044A,\
0x042E,0x0410,0x0411,0x0426,0x0414,0x0415,0x0424,0x0413,0x0425,0x0418,0x0419,0x041A,0x041B,0x041C,0x041D,0x041E,\
0x041F,0x042F,0x0420,0x0421,0x0422,0x0423,0x0416,0x0412,0x042C,0x042B,0x0417,0x0428,0x042D,0x0429,0x0427,0x042A};

static uint16_t iso88595_to_uni[256] = {\
0x0000,0x0001,0x0002,0x0003,0x0004,0x0005,0x0006,0x0007,0x0008,0x0009,0x000A,0x000B,0x000C,0x000D,0x000E,0x000F,\
0x0010,0x0011,0x0012,0x0013,0x0014,0x0015,0x0016,0x0017,0x0018,0x0019,0x001A,0x001B,0x001C,0x001D,0x001E,0x001F,\
0x0020,0x0021,0x0022,0x0023,0x0024,0x0025,0x0026,0x0027,0x0028,0x0029,0x002A,0x002B,0x002C,0x002D,0x002E,0x002F,\
0x0030,0x0031,0x0032,0x0033,0x0034,0x0035,0x0036,0x0037,0x0038,0x0039,0x003A,0x003B,0x003C,0x003D,0x003E,0x003F,\
0x0040,0x0041,0x0042,0x0043,0x0044,0x0045,0x0046,0x0047,0x0048,0x0049,0x004A,0x004B,0x004C,0x004D,0x004E,0x004F,\
0x0050,0x0051,0x0052,0x0053,0x0054,0x0055,0x0056,0x0057,0x0058,0x0059,0x005A,0x005B,0x005C,0x005D,0x005E,0x005F,\
0x0060,0x0061,0x0062,0x0063,0x0064,0x0065,0x0066,0x0067,0x0068,0x0069,0x006A,0x006B,0x006C,0x006D,0x006E,0x006F,\
0x0070,0x0071,0x0072,0x0073,0x0074,0x0075,0x0076,0x0077,0x0078,0x0079,0x007A,0x007B,0x007C,0x007D,0x007E,0x007F,\
0x0080,0x0081,0x0082,0x0083,0x0084,0x0085,0x0086,0x0087,0x0088,0x0089,0x008A,0x008B,0x008C,0x008D,0x008E,0x008F,\
0x0090,0x0091,0x0092,0x0093,0x0094,0x0095,0x0096,0x0097,0x0098,0x0099,0x009A,0x009B,0x009C,0x009D,0x009E,0x009F,\
0x00A0,0x0401,0x0402,0x0403,0x0404,0x0405,0x0406,0x0407,0x0408,0x0409,0x040A,0x040B,0x040C,0x00AD,0x040E,0x040F,\
0x0410,0x0411,0x0412,0x0413,0x0414,0x0415,0x0416,0x0417,0x0418,0x0419,0x041A,0x041B,0x041C,0x041D,0x041E,0x041F,\
0x0420,0x0421,0x0422,0x0423,0x0424,0x0425,0x0426,0x0427,0x0428,0x0429,0x042A,0x042B,0x042C,0x042D,0x042E,0x042F,\
0x0430,0x0431,0x0432,0x0433,0x0434,0x0435,0x0436,0x0437,0x0438,0x0439,0x043A,0x043B,0x043C,0x043D,0x043E,0x043F,\
0x0440,0x0441,0x0442,0x0443,0x0444,0x0445,0x0446,0x0447,0x0448,0x0449,0x044A,0x044B,0x044C,0x044D,0x044E,0x044F,\
0x2116,0x0451,0x0452,0x0453,0x0454,0x0455,0x0456,0x0457,0x0458,0x0459,0x045A,0x045B,0x045C,0x00A7,0x045E,0x045F};

static uint16_t cp866_to_uni[256] = {\
0x0000,0x0001,0x0002,0x0003,0x0004,0x0005,0x0006,0x0007,0x0008,0x0009,0x000a,0x000b,0x000c,0x000d,0x000e,0x000f,\
0x0010,0x0011,0x0012,0x0013,0x0014,0x0015,0x0016,0x0017,0x0018,0x0019,0x001a,0x001b,0x001c,0x001d,0x001e,0x001f,\
0x0020,0x0021,0x0022,0x0023,0x0024,0x0025,0x0026,0x0027,0x0028,0x0029,0x002a,0x002b,0x002c,0x002d,0x002e,0x002f,\
0x0030,0x0031,0x0032,0x0033,0x0034,0x0035,0x0036,0x0037,0x0038,0x0039,0x003a,0x003b,0x003c,0x003d,0x003e,0x003f,\
0x0040,0x0041,0x0042,0x0043,0x0044,0x0045,0x0046,0x0047,0x0048,0x0049,0x004a,0x004b,0x004c,0x004d,0x004e,0x004f,\
0x0050,0x0051,0x0052,0x0053,0x0054,0x0055,0x0056,0x0057,0x0058,0x0059,0x005a,0x005b,0x005c,0x005d,0x005e,0x005f,\
0x0060,0x0061,0x0062,0x0063,0x0064,0x0065,0x0066,0x0067,0x0068,0x0069,0x006a,0x006b,0x006c,0x006d,0x006e,0x006f,\
0x0070,0x0071,0x0072,0x0073,0x0074,0x0075,0x0076,0x0077,0x0078,0x0079,0x007a,0x007b,0x007c,0x007d,0x007e,0x007f,\
0x0410,0x0411,0x0412,0x0413,0x0414,0x0415,0x0416,0x0417,0x0418,0x0419,0x041a,0x041b,0x041c,0x041d,0x041e,0x041f,\
0x0420,0x0421,0x0422,0x0423,0x0424,0x0425,0x0426,0x0427,0x0428,0x0429,0x042a,0x042b,0x042c,0x042d,0x042e,0x042f,\
0x0430,0x0431,0x0432,0x0433,0x0434,0x0435,0x0436,0x0437,0x0438,0x0439,0x043a,0x043b,0x043c,0x043d,0x043e,0x043f,\
0x2591,0x2592,0x2593,0x2502,0x2524,0x2561,0x2562,0x2556,0x2555,0x2563,0x2551,0x2557,0x255d,0x255c,0x255b,0x2510,\
0x2514,0x2534,0x252c,0x251c,0x2500,0x253c,0x255e,0x255f,0x255a,0x2554,0x2569,0x2566,0x2560,0x2550,0x256c,0x2567,\
0x2568,0x2564,0x2565,0x2559,0x2558,0x2552,0x2553,0x256b,0x256a,0x2518,0x250c,0x2588,0x2584,0x258c,0x2590,0x2580,\
0x0440,0x0441,0x0442,0x0443,0x0444,0x0445,0x0446,0x0447,0x0448,0x0449,0x044a,0x044b,0x044c,0x044d,0x044e,0x044f,\
0x0401,0x0451,0x0404,0x0454,0x0407,0x0457,0x040e,0x045e,0x00b0,0x2219,0x00b7,0x221a,0x2116,0x00a4,0x25a0,0x00a0};

static inline uint16_t*
get_charmap(const char* encoding)
{
  uint16_t* charmap = 0;
  if (strcmp("koi8-r", encoding) == 0)
    charmap = koi8r_to_uni;
  else if (strcmp("cp1251", encoding) == 0)
    charmap = cp1251_to_uni;
  else if (strcmp("iso8859-5", encoding) == 0)
    charmap = iso88595_to_uni;
  else if (strcmp("cp866", encoding) == 0)
    charmap = cp866_to_uni;
  
  return charmap;
}

void draw_string(GC gc, const char* str, int n, int x, int y, DFBSurfaceTextFlags flags)
{
  uint16_t* charmap = get_charmap(fl_encoding);
  if (!charmap) {
    gc->surface->DrawString(gc->surface, str, n, x, y, flags);
    return;
  }
  const char *ch = str;
  int adv,p = n,_x = x;
  IDirectFBFont *font = get_surface_font(gc);
  for (ch = str; *ch && p--; ch++) {
    unsigned int uni = charmap[*ch];
    gc->surface->DrawGlyph(gc->surface, uni, _x, y, DSTF_LEFT); 
    font->GetGlyphExtents(font, uni, 0, &adv);
    _x += adv;
  }
  font->Release(font);
}

void _GetStringExtents(IDirectFBFont*, const char*, int, DFBRectangle*, DFBRectangle*);

#include <FL/fl_draw.H>

//FIXME: unfortunelly, freetype library does not print cyrillic encodings
//then I convert chars to UNICODE and draw the string glyph by glyph   

void _GetStringExtents(IDirectFBFont *font, const char* text,int bytes,
  DFBRectangle*   ret_logical_rect,
  DFBRectangle*   ret_ink_rect)
{
  uint16_t* charmap = get_charmap(fl_encoding);
  if (!charmap) {
    font->GetStringExtents(font, text, bytes, ret_logical_rect, ret_ink_rect);
    return;
  }
  const char *ch = text;
  int w = 0, adv = 0, p = bytes, y2 = 0;
  DFBRectangle rect;
  DFBRectangle sum_rect = {0, 0, 0, 0};
  for (ch = text; *ch && p--; ch++){
    font->GetGlyphExtents(font, charmap[*ch], &rect, &adv);
    if (rect.h && (sum_rect.y > rect.y))
      sum_rect.y = rect.y;
    if (y2 < rect.y + rect.h)
      y2 = rect.y + rect.h;
    w += adv;
  }
  sum_rect.w = w;
  sum_rect.h = y2 - sum_rect.y;
  if (ret_ink_rect)
    memcpy(ret_ink_rect, &sum_rect, sizeof(DFBRectangle));
}

void get_string_width(IDirectFBFont* font, const char* s, int n, int* width)
{
  if (!width)
    return;
  uint16_t* charmap = get_charmap(fl_encoding);
  if (charmap){
    const char *ch = s;
    int w = 0,adv = 0;
    for (ch = s; *ch && n--; ch++){
      font->GetGlyphExtents(font, charmap[*ch], 0, &adv);
      w += adv;
    }
    *width = w;
  }else
    font->GetStringWidth(font, s, n, width);
}

/***********************************************************************************/

//make mixed color 
ulong
average(ulong rgb1,ulong rgb2,double weight)
{
  ulong r = (uchar)(((uchar)(rgb1 >> 16)) * weight + ((uchar)(rgb2 >> 16)) * (1 - weight));
  ulong g = (uchar)(((uchar)(rgb1 >>  8)) * weight + ((uchar)(rgb2 >>  8)) * (1 - weight));
  ulong b = (uchar)(((uchar)(rgb1 >>  0)) * weight + ((uchar)(rgb2 >>  0)) * (1 - weight));
  return (r << 16) | (g << 8) | b;
}
//make mixed color 
ushort
average(ushort rgb1, ushort rgb2, double weight)
{
  ushort r = (ushort)(RGB16_RED  (rgb1) * weight + RGB16_RED  (rgb2) * (1 - weight) + .5);
  ushort g = (ushort)(RGB16_GREEN(rgb1) * weight + RGB16_GREEN(rgb2) * (1 - weight) + .5);
  ushort b = (ushort)(RGB16_BLUE (rgb1) * weight + RGB16_BLUE (rgb2) * (1 - weight) + .5);
  return PIXEL_RGB16(r, g ,b);
}

uchar
average(uchar c1, uchar c2,double weight)
{
  uchar r1, r2, g1, g2, b1, b2;
  get_rgb_color(c1, r1, g1, b1);
  get_rgb_color(c2, r2, g2, b2);
  ulong r = (uchar)(r1 * weight + r2 * (1 - weight));
  ulong g = (uchar)(g1 * weight + g2 * (1 - weight));
  ulong b = (uchar)(b1 * weight + b2 * (1 - weight));
  return get_index_color(r, g, b);
}

/////////////////////////////////////////////////////////////////////////////////
//implementation of class 'direct_draw'

direct_draw::direct_draw(GC gc)
{
  begin_draw(gc);
}

direct_draw::direct_draw(GC gc, uchar* bm)
{
  begin_draw(gc);
  mbm = bm;
  mbmw = (mw + 7) / 8;
}

void direct_draw::begin_draw(GC gc)
{
  mgc = gc;
  if (mgc){
    mpixf = get_pixel_format(mgc);
    mbpp = DFB_BYTES_PER_PIXEL(mpixf);
    mbuf = (uchar*)lock_surface(mgc, DSLF_WRITE, &mpitch);
    get_size(mgc, &mw, &mh);
    mcolor = dfb_color_to_pixel((DFBSurfacePixelFormat)mpixf,mgc->r,mgc->g,mgc->b);
  }
  mbm=0;
}

void direct_draw::end_draw()
{
  if (mgc){
    mgc->surface->Unlock(mgc->surface);
      mgc=0;
  }
}

direct_draw::~direct_draw()
{
  end_draw();
}

void direct_draw::setcolor(u8 r, u8 g, u8 b, u8 a)
{
  mcolor = dfb_color_to_pixel((DFBSurfacePixelFormat)mpixf, r, g, b);
}

int direct_draw::getpixel(int x, int y) const
{
  if (x >= 0 && y >= 0 &&  x <= mw && y <= mh){
    switch (mbpp) {
    case 1:
      return ((uchar*)mbuf)[y * mpitch + x];
        case 2:
            return ((u16*)mbuf)[y * mpitch / 2 + x];
    case 4:
      return ((ulong*)mbuf)[y * mpitch / 4 + x];
    }
  }
  return -1;
}

void direct_draw::setbit(int x,int y){
  if (mbm){
    uchar *pos=mbm + y*mbmw + x/8;
    *pos |= 1<<(x&7);
  }
}

int direct_draw::getbit(int x,int y){
  if (mbm){
    uchar *pos=mbm+y*mbmw + x/8;
    return (*pos & (1<<(x&7)))!=0;
  }
  return 0;
}

void direct_draw::setpixel(int x, int y, double br)
{
  if (x >= 0 && y >= 0 && x <= mw && y <= mh){
    switch (mbpp){
    case 1: {
      uchar *pos=mbuf+y*mpitch+x;
      if (dash(mgc) /*&& PixInReg(mgc, x ,y)*/){
        uchar col = average(mcolor, *pos, br);
        *pos = col;
        setbit(x,y);
      }
      break;
    }
    case 2: {
      u16 *pos = &((u16*)mbuf)[y*mpitch/2 + x];
      if (dash(mgc) /*&& PixInReg(mgc, x ,y)*/){
        u16 col = average(mcolor, LOAD_U16_VO(pos), br);
        STORE_U16_VO(pos, col);
        setbit(x,y);
      }
      break;
    }
    case 4: {
      ulong *pos = &((ulong*)mbuf)[y*mpitch/4 + x];
      if (dash(mgc) /*&& PixInReg(mgc, x ,y)*/){
        ulong col = average(mcolor, (ulong)LOAD_U32_VO(pos), br);
        STORE_U32_VO(pos, col);
        setbit(x,y);
      }
      break;
    }
    }
  }
}

void direct_draw::setpixel(int x, int y)
{
  if (x >= 0 && y >= 0 &&  x <= mw && y <= mh) {
    if (dash(mgc) /*&& PixInReg(mgc, x , y)*/) {
      switch (mbpp) {
      case 1: {
        uchar *pos = &((uchar*)mbuf)[y*mpitch + x];
        *pos = mcolor;
        setbit(x,y);
        break;
      }
      case 2: {
        u16 *pos = &((u16*)mbuf)[y*mpitch/2 + x];
        STORE_U16_VO(pos, mcolor);
        setbit(x,y);
        break;
      }
      case 4: {
        ulong *pos = &((ulong*)mbuf)[y*mpitch/4 + x];//FIXME pitch measures in bytes or in pixels ?
        STORE_U32_VO(pos, mcolor);
        setbit(x,y);
        break;
      }
      }
    }
  }
}

void direct_draw::setpixel(int x, int y, int color){
  if (x >= 0 && y >= 0 &&  x <= mw && y <= mh){
    if (dash(mgc) /*&& PixInReg(mgc, x , y)*/){
      switch (mbpp) {
      case 1: {
        uchar *pos = &((uchar*)mbuf)[y*mpitch + x];
        *pos = (uchar)color;
        setbit(x,y);
        break;
      }
      case 2:{
        u16 *pos = &((u16*)mbuf)[y*mpitch/2 + x];
        STORE_U16_VO(pos, color);
        setbit(x,y);
        break;
      }
      case 4:{
        ulong *pos = &((ulong*)mbuf)[y*mpitch/4 + x];
        STORE_U32_VO(pos, color);
        setbit(x,y);
        break;
      }
      }
    }
  }
}
/*
void direct_draw::setrow_x2(int _x1, int _x2, int y){
  
  int x1=_x1, y1=y, x2=_x2, y2=y, d=1,x;
  if (x1 > x2){
    x2 = _x1;
    x1 = _x2;
    d=-1;
  }
    
  if (d<0) x = x2;
    else x = x1;
    
  int w = x2-x1+1;
  if (mbpp==1) {  
    uchar *pos = &((uchar*)mbuf)[y*mpitch + x];
    for (int i = 0; i < w; i++){
      if (dash())
        *pos = mcolor;
      pos+=d;
    }
  }else if (mbpp==4){
    ulong *pos = &((ulong*)mbuf)[y*mpitch/4 + x];//FIXME pitch ?
    for (int i = 0; i < w; i++){
      if (dash())
        *pos = mcolor;
      pos+=d;
    }
  }
}
*/
//fill row limited pixels with border_color
void direct_draw::fillrow(int& left, int& right,int y,ulong border_color)
{
  if (mbpp==1){
    uchar *pos = &((uchar*)mbuf)[y*mpitch+right];
    uchar *pos2=pos;
    do
      *pos++ = mcolor;
    while((*pos!=border_color)&&(++right<mw));
    
    while((*--pos2!=border_color)&&(--left>=0))
      *pos2 = mcolor;
  }else if (mbpp == 4){
    ulong *pos = &((ulong*)mbuf)[y * mpitch / 4 + right];
    ulong *pos2=pos;
    do  *pos++ = mcolor; while((*pos != border_color)&&(++right<mw));
    
    while((*--pos2!=border_color)&&(--left>=0))  *pos2 = mcolor;
  }
}

//fill row using bitmap 
void direct_draw::fillrow(int& left, int& right, int y){//left==right !!!
  if (!mbm)//it need bitmap
    return;
  uchar *bmpos=mbm+y*mbmw + right/8;
  uchar *bmpos2=bmpos;
  if (mbpp==1){  
    uchar *pos = &((uchar*)mbuf)[y*mpitch + right];
    uchar *pos2=pos;
    do{
      if (((right&7)==0)&&(*bmpos==0)&&(right+8<mw)){
        memset(pos,mcolor,8);
        *bmpos++=255;
        right+=8;
        pos+=8;
        _TRACE("right %d\n",right);        
        continue;
      }else{
        *pos++ = mcolor;
        *bmpos |= 1<<(right&7);
        if ((right&7)==7) bmpos++;
        right++;
      }
    }while(!(*bmpos & (1<<(right&7)))&&right<mw);
    
    while(1){
      if ((--left&7)==7){
        bmpos2--;
        if (*bmpos2==0 && left>=8){
          *bmpos2 = 255;
          left-=7;
          pos2-=8;
          memset(pos2,mcolor,8);
          _TRACE("left %d\n",left);
          continue;
        }
      }
      if (!(*bmpos2 & (1<<(left&7)))&&(left>=0)){
        *(--pos2) = mcolor;
        *bmpos2 |= 1<<(left&7);
      }else
        break;
    }
  }else if (mbpp==4){
    ulong *pos = &((ulong*)mbuf)[y*mpitch/4 + right];
    ulong *pos2=pos;
    int i;
    do{
      if (((right&7)==0)&&(*bmpos==0)&&(right+8<mw)){
        for (i=0;i<8;i++)*pos++=mcolor;
        *bmpos++=255;
        right+=8;
        _TRACE("right %d\n",right);        
        continue;
      }else{
        *pos++=mcolor;
        *bmpos |= 1<<(right&7);
        if ((right&7)==7) bmpos++;
        right++;
      }
    }while(!(*bmpos & (1<<(right&7)))&&right<mw);
    
    while(1){
      if ((--left&7)==7){
        bmpos2--;
        if (*bmpos2==0 && left>=8){
          *bmpos2 = 255;
          left-=7;
          for (i=0;i<8;i++)*(--pos2)=mcolor;
          _TRACE("left %d\n",left);
          continue;
        }
      }
      if (!(*bmpos2 & (1<<(left&7)))&&(left>=0)){
        *(--pos2) = mcolor;
        *bmpos2 |= 1<<(left&7);
      }else
        break;
    }
  }
}
  
void direct_draw::setrow(int x, int y, int w){
      
  int x1=x,x2=x+w+((w<0)?1:-1),/* y2=y, y1=y,*/ d=1;
        
  if (x2 > x1){
    w=x2-x1+1;
  }else{
    w=x1-x2+1;
    d=-1;
  }
  x = x1;
  switch (mbpp){
        case 1:{  
      uchar *pos = &((uchar*)mbuf)[y*mpitch+x];
      for (int i = 0; i < w; i++){
        if (dash(mgc))
          *pos = mcolor;
        pos+=d;
      }
            break;
        }
        case 2:{
            u16 *pos = &((u16*)mbuf)[y*mpitch/2 + x];
            for (int i = 0; i < w; i++){
                if (dash(mgc))
                    STORE_U16_VO(pos, mcolor);
                pos+=d;
            }
            break;
        }
        
        case 4:{
      ulong *pos = &((ulong*)mbuf)[y*mpitch/4 + x];
      for (int i = 0; i < w; i++){
        if (dash(mgc))
          STORE_U32_VO(pos, mcolor);
        pos+=d;
      }
            break;
    }
    }
}

void direct_draw::setcol(int x, int y, int h){
  int /*x1=x,x2=x,*/ y1=y,  y2=y+h+((h<0)?1:-1),d=1;
        
  if (y2 > y1){
    h=y2-y1+1;
  }else{
    h=y1-y2+1;
    d=-1;
  }
  y=y1;
  switch (mbpp) {
       case 1:{  
      uchar *pos = &((uchar*)mbuf)[y*mpitch + x];
      int pitch = mpitch*d;
      for (int i = 0; i<h; i++){
        if (dash(mgc)) *pos = mcolor;
        pos+=pitch;
      }
            break;
    }
        case 2:{
            int pitch = mpitch/2;
            u16 *pos = &((u16*)mbuf)[y*pitch + x];
            pitch *= d;
            for (int i = 0; i<h; i++){
                if (dash(mgc))
                    STORE_U16_VO(pos, mcolor);
                pos+=pitch;
            }
            break;
        }   
        case 4:{
      int pitch = mpitch/4;
      ulong *pos = &((ulong*)mbuf)[y*pitch + x];
      pitch *= d;
      for (int i = 0; i<h; i++){
        if (dash(mgc)) 
                    STORE_U32_VO(pos, mcolor);
        pos+=pitch;
      }
            break;
    }  
    }
}

//////////////////////////////////////////////////////////
// draw simple line using the direct_draw class

void draw_line(GC gc, DFBRegion *line){
    int x,y,dx,dy,sx = 1,sy = 1,z,e,i;
    bool ch;
    
    DFBRegion clip;
    get_clip(gc, &clip);
    if (!dfb_clip_line(&clip, line))
        return;
    
    x = line->x1;
    y = line->y1;
    dx = line->x2 - line->x1;
    dy = line->y2 - line->y1;
    
    if (dx < 0){
        dx = -dx;sx = -1;
    }
    if (dy < 0){
        dy = -dy;sy = -1;
    }
    direct_draw dd(gc);
          
    if (dx==0 && dy==0){//only one point
        dd.setpixel(x, y);
        dd.end_draw();
        return;
    }
    if (dy == 0){//horizontal line
     dd.setrow(sx>0 ? x :line->x2, y, dx);
     dd.end_draw();
     return;
    }     
    if (dx == 0){//vertical line
     dd.setcol(x , sy>0 ? y :line->y2, dy);
     dd.end_draw();
     return;
    }
    if (dy>dx){
        z = dx;dx = dy;dy = z;
        ch = true;
    } else{
        ch = false;
    }
    e = 2*dy-dx;
    i = 1;
    do{
        dd.setpixel(x, y);
        while(e>=0){
            if (ch){
                x = x+sx;
            }else{
                y = y+sy;
            }
            e = e-2*dx;
        }
        if (ch){
            y = y+sy;
        }else{
            x = x+sx;
        }
        e = e+2*dy;
        i = i+1;
    }
    while(i<=dx);
    dd.setpixel(x, y);
    dd.end_draw();
}

///////////////////////////////////////////////////////////////////////////
// smooth and filter functions
// objects to change....

// masks for filtering 
static int
   Mask0[]= {1,1, 1,1 },  // 2 X 2
   Mask1[]= {1,1,1, 1,1,1, 1,1,1 },//3 X 3
   Mask2[]= {1,1,1, 1,2,1, 1,1,1 },//3 X 3
   Mask3[]= {1,2,1, 2,4,2, 1,2,1 },//3 X 3
   Mask4[]= {0,2,0, 2,4,2, 0,2,0 },//3 X 3
   Mask5[]= {1,1,1,1,
             1,1,1,1,
             1,1,1,1,
             1,1,1,1 },//4 X 4
   Mask6[]= {1,2, 3, 4, 3,2,1,
             2,4, 6, 8, 6,4,2,
             3,6, 9,12, 9,6,5,
             4,8,12,16,12,8,4,
             3,6, 9,12, 9,6,5,
             2,4, 6, 8, 6,4,2,
             1,2, 3, 4, 3,2,1 },//7 X 7
   Mask_ln[]= {2, 3, 3, 3, 3, 4, 7},    /* size of mask   */
   Mask_vl[]= {4, 9,10,16,12,16,256},   /* sum of all items */
   *Mask_bg[]={                         /* mask by index  */
      Mask0,Mask1,Mask2,Mask3,Mask4,Mask5,Mask6
   };


#include <math.h>

static double myfrac(double x){
    return x-floor(x);
}

// draw smooth line
void draw_smooth_line(GC gc, double x1, double y1, double x2, double y2){
    double grad, xd, yd,xgap,xend,yend,yf,brightness1,brightness2;
    int x;
    int ix1;
    int ix2;
    int iy1;
    int iy2;
    bool wasexchange,invert=0;
    double tmpreal;
   
    xd = x2-x1;
    yd = y2-y1;
    if (xd==0&&yd==0){
        return;
    }
    if (fabs(xd)>fabs(yd)){
        wasexchange = false;
    }else{
        wasexchange = true;
        tmpreal = x1;x1 = y1;y1 = tmpreal;
        tmpreal = x2;x2 = y2;y2 = tmpreal;
        tmpreal = xd;xd = yd;yd = tmpreal;
    }
    if (x1>x2){
        invert=1;
        tmpreal = x1;x1 = x2;x2 = tmpreal;
        tmpreal = y1;y1 = y2;y2 = tmpreal;
        xd = x2-x1;
        yd = y2-y1;
    }
    grad = yd/xd;
    xend = floor(x1+0.5);
    yend = y1+grad*(xend-x1);
    xgap = 1-myfrac(x1+0.5);
    ix1 = (int)floor(x1+0.5);
    iy1 = (int)floor(yend);
    brightness1 = (1-myfrac(yend))*xgap;
    brightness2 = myfrac(yend)*xgap;
    
    direct_draw dd(gc);
          
    if (wasexchange){
        dd.setpixel(iy1, ix1, brightness1);
        dd.setpixel(iy1+1, ix1, brightness2);
    }else{
        dd.setpixel(ix1, iy1, brightness1);
        dd.setpixel(ix1, iy1+1, brightness2);
    }
    yf = yend+grad;
    xend = floor(x2+0.5);
    yend = y2+grad*(xend-x2);
    xgap = 1-myfrac(x2/*-0.5*/);
    ix2 = (int)floor(x2+0.5);
    iy2 = (int)floor(yend);
    brightness1 = (1-myfrac(yend))*xgap;
    brightness2 = myfrac(yend)*xgap;
    if (wasexchange){
        dd.setpixel(iy2, ix2, brightness1);
        dd.setpixel(iy2+1, ix2, brightness2);
    }else{
        dd.setpixel(ix2, iy2, brightness1);
        dd.setpixel(ix2, iy2+1, brightness2);
    }
    for (x = ix1+1; x <= ix2-1; x++){
        brightness1 = 1-myfrac(yf);
        brightness2 = myfrac(yf);
        if (wasexchange){
            dd.setpixel((int)floor(yf), x, brightness1);
            dd.setpixel((int)floor(yf)+1, x, brightness2);
        }else{
            dd.setpixel(x, (int)floor(yf), brightness1);
            dd.setpixel(x, (int)floor(yf)+1, brightness2);
        }
        yf = yf+grad;
    }
    dd.end_draw();
}

// draw smooth edge
void draw_edge_(GC gc, double x1, double y1, double x2, double y2,uchar* bm,int fill){
    double grad, xd, yd,xgap,xend,yend,yf,brightness1,brightness2;
    int x,ix1,ix2,iy1,iy2;
    bool wasexchange;
    double tmpreal,vx1=x1,vy1=y1;
    
    static double prev_x2=0,prev_y2=0;
    bool is_vertex=false;
    
  if ((x1 == prev_x2)&&(y1 == prev_y2))
    is_vertex=true;
  
  prev_x2=x2;prev_y2=y2;
     
    xd = x2-x1;
    yd = y2-y1;
    if (xd==0&&yd==0){
        return;
    }
    
    if (fabs(xd)>fabs(yd)){
        wasexchange = false;
    }else{
        wasexchange = true;
        tmpreal = x1; x1 = y1;  y1 = tmpreal;
        tmpreal = x2; x2 = y2;  y2 = tmpreal;
        tmpreal = xd; xd = yd;  yd = tmpreal;
  }
  bool invert=false;
  if (x1>x2){
      tmpreal = x1;x1 = x2;x2 = tmpreal;
        tmpreal = y1;y1 = y2;y2 = tmpreal;
        xd = x2-x1;
        yd = y2-y1;
    }else{
      invert=true;
  }
  if (!wasexchange)
    invert = !invert;
    
    grad = yd/xd;
    xend = floor(x1+.5);
    yend = y1+grad*(xend-x1);
    xgap = 1-myfrac(x1+.5);
    ix1 = (int)floor(x1+.5);
    iy1 = (int)floor(yend);

     brightness1 = (1-myfrac(yend))*xgap;
  brightness2 = myfrac(yend)*xgap;
    
  _TRACE("x1 %.2lf y1 %.2lf x2 %.2lf y2 %.2lf xd %.2lf yd %.2lf xend %.2lf yend %.2lf ix1 %d iy1 %d\n",x1,y1,x2,y2,xd,yd,xend,yend,ix1,iy1); 
    direct_draw dd(gc,bm);
    
    if (wasexchange){
      if (!invert)
          dd.setpixel(iy1, ix1, brightness1);
         if (invert)
          dd.setpixel(iy1+1, ix1, brightness2);
    }else{
      if (!invert)
          dd.setpixel(ix1, iy1, brightness1);
        if (invert)  
          dd.setpixel(ix1, iy1+1, brightness2);
    }
    //FIXME I put bit to stop a hole when it's needed
    if (is_vertex){
        int _x=(int)vx1,_y=(int)(vy1+.5);
       if (vx1 > x1)
         _x=(int)(vx1+1);
       _TRACE(" _x %d, _y %d\n", _x,_y);  
       if (!dd.getbit(_x-1,_y))
         if (!dd.getbit(_x+1,_y))
           if (!dd.getbit(_x,_y-1))
             if (!dd.getbit(_x,_y+1))
               dd.setbit(_x,_y);
  }
    yf = yend+grad;
    xend = floor(x2+.5);
    yend = y2+grad*(xend-x2);
    xgap = 1-myfrac(x2-.5);
    ix2 = (int)floor(x2+.5);
    iy2 = (int)floor(yend);
    brightness1 = (1-myfrac(yend))*xgap;
    brightness2 = myfrac(yend)*xgap;
    if (wasexchange){
      if (!invert)
          dd.setpixel(iy2, ix2, brightness1);
        if (invert)  
          dd.setpixel(iy2+1, ix2, brightness2);
    }else{
      if (!invert)
          dd.setpixel(ix2, iy2, brightness1);
        if (invert)
          dd.setpixel(ix2, iy2+1, brightness2);
    }
    for (x = ix1+1; x<=ix2-1; x++){
        brightness1 = 1-myfrac(yf);
        brightness2 = myfrac(yf);
        if (wasexchange){
          if (!invert)
              dd.setpixel((int)floor(yf),x,brightness1);
            else  
              dd.setpixel((int)floor(yf)+1, x, brightness2);
        }else{
          if (!invert)
              dd.setpixel(x,(int)floor(yf),brightness1);
            else  
              dd.setpixel(x, (int)floor(yf)+1, brightness2);
        }
        yf = yf+grad;
    }
  dd.end_draw();
}

#if 0

#define Pix_C 64
//#define Pix_V 64 

#define my_int(f)(f<.0)?(int)(f-.5):(int)(f+.5)

// old function which takes an integer co-ordinates 
void draw_edge(GC gc,float x1,float y1,float x2,float y2,uchar* bm,int fill){
   float dx, dy;
   int sx, sy,swap;
   float incr1=0, incr2=0;
   float E, kl;              /* current error  */
   float Emax;          /* max error     */
   float color_tek;      /* current color  */
   int  xt;
   if ((x1<x2 && y1<y2)||(x1>x2 && y1>y2)){
       float tmp=x1;x1=x2;x2=tmp;
       tmp=y1;y1=y2;y2=tmp;
   }
/* calc deltas and steps */
  sx= 0;
  if ((dx= (x2-x1))<0){dx=-dx;--sx;} else if (dx>0) ++sx;
  sy= 0;
  if ((dy= (y2-y1))<0){dy=-dy;--sy;} else if (dy>0) ++sy;
/* check sloping */
  swap= 0;
  if ((kl=dx)<(E=dy)){dx=E; dy=kl; kl=E; ++swap;}
    
    float t= 0;
    if (dx)
     t= dy/dx;
            
  Emax= 1-t;
    
    E= 0.5-((1-myfrac(y1)) - t*myfrac(x1)); /* set error */
    _TRACE("%.2lf %.2lf \n", myfrac(x1),myfrac(y1));
                             
//  s_max=incr2-incr1;    
      
  direct_draw dd(gc, bm);
  color_tek = 1;         
  if (dx)
    color_tek=(dy/dx)/2;
     
  int xn=(int)(x1),yn=(int)(y1),xk=(int)(x2+.5),yk=(int)(y2+.5);
    _TRACE("xn %d yn %d xk %d yk %d t %.4f E %.4f Emax %.4f\n",xn,yn,xk,yk,t,E,Emax);
  dd.setpixel(xn, yn,color_tek);      /* first pixel */
  if (swap && fill){
    xt=xn;
    if (xn < xk){
         while(++xt <= xk) dd.setpixel(xt,yn,1.0);
       }else{
      while(--xt >= xk) dd.setpixel(xt,yn,1.0);
      }
   }
    while(--kl >= 0){
      if (E >= Emax) {
         if (swap){
            xn+= sx;
            yn+= sy;
        }else{
            if (fill){ 
             xt=xn;
          if (xn < xk){while(++xt <= xk)dd.setpixel(xt,yn,1.0);}
          else{while(--xt >= xk)dd.setpixel(xt,yn,1.0);}
           }
             yn+= sy;
            xn+= sx;
         }
         E -= Emax;
      }else{
          if (swap)
            yn+= sy;
          else
            xn+= sx;
          
          E += t;
      }
     //     color_tek= 1;
          
    //      if (dx) 
         color_tek = E;
                
      if (swap)
      dd.setpixel(xn,yn,1-color_tek);     /* current pixel */
      else{
      dd.setpixel(xn,yn,color_tek);     /* current pixel */
        _TRACE("(%d %d) %.3f\n",xn,yn,E);
      }
    /* filling */
     if (swap&&fill){
             xt=xn;
         if (xn < xk) {while(++xt <= xk) dd.setpixel(xt,yn,1.0);}
         else {while(--xt >= xk) dd.setpixel(xt,yn,1.0);}
        }
   }
}  /* draw_edge */

void draw_edge(float x1,float y1,float x2,float y2){
    draw_edge(fl_gc,x1,y1,x2,y2,0,1);
}

int _smooth = 0;

void drawrow(GC gc,double x1,double x2,double y){
  if (_smooth){
     draw_smooth_line(gc,x1,y,x2,y);
  }else{
    gc->DrawLine(gc,(int)(x1+.5),(int)(y+.5),(int)(x2+.5),(int)(y+.5));
  }
}
#endif

void fl_filter_widget(Fl_Widget *wdg,int mask){
  fl_filter_offscreen(fl_gc, wdg->x(), wdg->y(), wdg->w(), wdg->h(),mask);
}

/*-------------------------- fl_filter_offscreen */

void fl_filter_offscreen(GC gc,int x,int y,int w,int h,int mask)
{
  
   int  *pi;            
   int  pixel;          
   int  *cur_mask,         
        mask_h,mask_w,  
        x_center,y_center,
        Mask_sum,Xk,Yk,             
        s, sr, sg, sb,ii, jj,
       Xt, Yt;
       
    if (!mask) return; // zero mask - do nothing
    if (!gc) // if offscreen is NULL use current surface
      gc=fl_gc;
   cur_mask=  Mask_bg[--mask];            
  mask_h= mask_w= Mask_ln[mask];    
  x_center= mask_w / 2;
  y_center= mask_h / 2;
  Mask_sum= Mask_vl[mask];
  //makes a small RGB surface 
  GC surface = create_mem_gc(w,h,DSPF_RGB32);
  DFBRectangle rect_d;
  rect_d.x = x;rect_d.y = y;
  rect_d.w = w;rect_d.h = h;
  //copy the image and convert it to RGB if needed
  blit(surface,gc,&rect_d,0,0); 
  int pitch_d;
  uchar *buffer = (uchar*)lock_surface(surface, DFBSurfaceLockFlags(DSLF_WRITE|DSLF_READ),&pitch_d);
  uchar *src = (uchar*)Xmalloc(pitch_d*h); 
  memcpy(src, buffer, pitch_d*h);

/* maximal positions */
  Xk= w-mask_w;
  Yk= h-mask_h;

  ulong *ppix = 0;
  int ppix_inc = (pitch_d/4-mask_w);
  for (Yt=0 /*y1_src*/; Yt<=Yk; ++Yt) {
    for (Xt=0/*x1_src*/; Xt<=Xk; ++Xt) {
      pi= cur_mask; sr=0; sg=0; sb=0;   /* RGB*/
      ppix = &((ulong*)src)[Xt+Yt*pitch_d/4];
      for (ii=0; ii<mask_h; ++ii){
        for (jj=0; jj<mask_w; ++jj) {
          pixel= *ppix++;
                   s= *pi++; 
                sr+= (s*((pixel>>16)&255));  
                   sg+= (s*((pixel>>8)&255));   
                   sb+= (s*((pixel)&255));      
              }
              ppix+=ppix_inc;
           }
           sr /= Mask_sum; sg /= Mask_sum; sb /= Mask_sum;
      ii= PIXEL_RGB32(sr,sg,sb);
      ((ulong*)buffer)[(/*x1_dst+*/(Xt/*-x1_src*/)+x_center) + (/*y1_dst+*/(Yt/*-y1_src*/)+y_center)*pitch_d/4]=ii;
     }
  }
    unlock_surface(surface);
     Xfree(src);
  rect_d.x=rect_d.y=0;
   blit(gc,surface,&rect_d,x,y);
   release_gc(surface);
}  /* V_fltr0 */

/*----------------------------------------------------- V_FAST */

#define MAX_GOR 2048  /* resolution by X */
#define MAX_VER 2048  /* resolution by Y */
#define MAX_STK 8192  /* size of stack */

static int gor_max= MAX_GOR;
static int ver_max= MAX_VER;
static int stk_max= MAX_STK;
static int *pi_stk, *pn_stk;   
static int xtek, ytek;
static int stklen;

static direct_draw* _dd=0;

/***********************************************************/
static uchar *_bm=0;  //bitmap 
static int _bm_w=0;    //width of bitmap in bytes

inline bool getbit(int x,int y){
  return  (bool)(*(_bm + y*_bm_w +x/8) & (1<<(x&7)));
}

inline void setbit(int x,int y){
  *(_bm+y*_bm_w + x/8) |= (1<<(x&7));
}

inline void clearbit(int x,int y){
    *(_bm+y*_bm_w + x/8) &= ~(1<<(x&7));
}

inline uchar getbyte(int x,int y){
  return *(_bm+y*_bm_w+x/8);
}

inline void setbyte(int x,int y){
  *(_bm+y*_bm_w+x/8)=255;
}
/************************************************************/

/*---------------------------------------------------- Pop_Stk */
static int  Pop_Stk (){
  int  ret;
  ret= 0;
  if (pi_stk <= pn_stk) ++ret; else {
    ytek= *--pi_stk;  xtek= *--pi_stk;
  }
  _TRACE("pop %d %d\n",xtek,ytek);
  return (ret);
}  /* Pop_Stk */

/*--------------------------------------------------- Push_Stk */
static int  Push_Stk (int x,int y){
  int glu;
  _TRACE("push %d %d\n",x,y);
  if ((glu= pi_stk - pn_stk) >= stk_max) x= -1; else {
    *pi_stk++= x;  *pi_stk++= y; x= 0;
    if (glu > stklen) stklen= glu;
  }
  return (x);
}  /* Push_Stk */

static int Search (int yt,int left,int right,int border_color,int fill_color){
  int x, ret,pix;
  _TRACE(" search y=%d  left=%d, l= %d\n",yt,left,right-left);
  ret= 0;
    if (yt>=ver_max || yt<0)
        return 0;
  if (right>gor_max) right=gor_max;
  while (left <= right) {
    x=left;
  /* find right pixel */
    if (!_bm){
      while((pix=_dd->getpixel(x,yt))!=border_color && pix!=fill_color && x<right)
        ++x;
      if (x != left) {          /* found */
        ++ret;
        if (x != right || pix == border_color || pix == fill_color) --x;
        Push_Stk(x, yt);
      }
    }else{
      while(!getbit(x,yt) && x<right){
        while(((x&7)==0) && !getbyte(x,yt) && ((x+8)<right)) 
          x+=8;  
        x++;
      }
        if (x!=left){          /* found */
        ++ret;
        if (x!=right||getbit(x,yt))--x;
        Push_Stk (x, yt);
      }
    }
  /* continue scan line */
    left=x;
    if (!_bm){
         while(((pix=_dd->getpixel(x,yt))==border_color||pix==fill_color) && x<right)
             ++x;
    }else{
      while(getbit(x,yt) && x<right){
        while(((x&7)==0) && (getbyte(x,yt)==255) && (x+8)<right) 
          x+=8;
         ++x;
      }
    }
    if (x==left)++x;
    left=x;
  }
  return (ret);
}  /* Search */

// forward declaration 
int fill_(GC gc,int border_color,int fill_color,int x_isx,int y_isx,uchar* bm);

// filling a figure limited with border_color 
int fill(GC gc,int border_color,int fill_color,int start_x,int start_y){
  return fill_(gc,border_color,fill_color,start_x,start_y,0);
}
//filling a figure using bitmask 
int fill(GC gc,int fill_color,int start_x,int start_y,uchar* bm){
  return fill_(gc,fill_color,fill_color,start_x,start_y,bm);
}
// main function of filling   
int fill_(GC gc,int border_color,int fill_color,int x_isx,int y_isx,uchar* bm){
  int  left,right,w,h;
  int  ret=0;
  
     _bm=bm;
  get_size(gc,&w,&h);
     _bm_w=(w+7)/8;
     
  gor_max=w;
  ver_max=h;
       
  ret= 0;
/* stack init */
  if ((pn_stk=(int *)Xmalloc(stk_max+1)) == NULL) {
    --ret;  goto all;
  }
  pi_stk= pn_stk;

  Push_Stk (x_isx, y_isx);     /*    */
  _dd = new direct_draw(gc,bm);
  _dd->setcolor(fill_color>>16, fill_color>>8, fill_color, fill_color>>24); 
  
/* cycle while stack is not empty */
  while (pi_stk > pn_stk){

/* load next position */
    if (Pop_Stk ()) {ret=2; break; }
    if ((xtek<0)||(xtek>w)||ytek<0||ytek>h)
      break;
      
    left=right=xtek;   /* start position */
    _TRACE("start x %d y %d\t", xtek,ytek);
    if (!_bm)
      _dd->fillrow(left,right,ytek,border_color);  
        else
      _dd->fillrow(left,right,ytek);
          
        _TRACE("left %d right %d\n",left,right);
        --right;  
    ++left;

    /*search in next line*/
    Search (++ytek, left, right, border_color, fill_color);

      ytek-=2;
      /*search in prev line*/
    Search (ytek, left, right, border_color, fill_color);
     }
     Xfree (pn_stk);
all:
  _dd->end_draw();
  delete _dd;
  _dd=0;
     return (ret);
} /*end of fill_*/

// for debug only
void print_bitmap(uchar* bm,int w,int h){
  int bw = (w+7)/8;
    for (int i=0;i<h;i++){
    for (int j=0;j<bw;j++){
      uchar b=bm[i*bw+j];  
      for (int l=0;l<8;l++)
      if (b&(1<<l))TRACE("X");
      else TRACE(" ");
    }
    TRACE("\n");
  }
}

#include <FL/fl_draw.H>

/* draw SIMPLE polygon with smooth edges 
   outside edge lays on left hand side
   it's needed to discribe points in clockwise direction
   
 1          2
  *--->-----* 
  |######### \
  |########## \
  ^########### V
  |############ \
  |############# \
  *-------<-------* 
 0                 3
 
*/

void SmoothPoly(GC gc, int count, POINT_f *pt){
//    return;
    int i,min_x=10000,min_y=10000,max_x=0,max_y=0,sw,sh,size,x,y;
/* find borders of the polygon */
    for (i=0; i<count;i++){
      if (pt[i].x>max_x) max_x=(int)(pt[i].x+.5);
      if (pt[i].x<min_x) min_x=(int)(pt[i].x);
      if (pt[i].y>max_y) max_y=(int)(pt[i].y+.5);
      if (pt[i].y<min_y) min_y=(int)(pt[i].y);
    }
    DFBRectangle rect;
    rect.x=min_x;rect.y=min_y;rect.w=max_x-min_x+2;rect.h=max_y-min_y+2;
    GC subs=0;
/* clip the surface */
  subs = create_gc(gc, &rect);
    if (!subs) 
        return;
  get_size(subs,&sw,&sh);
  size = sh*((sw+7)/8);
/* allocate an empty bitmask */
  uchar* bm = (uchar*)Xcalloc(size,1);
/* draw the borders of polygon and create bitmask */
    for (i=0; i<count;i++){
        pt[i].x-=min_x;
        pt[i].y-=min_y;
        _TRACE("x %.1f y %.1f \n",pt[i].x,pt[i].y);
    }
  for (i=1;i<count;i++){
      draw_edge_(subs,pt[i-1].x,pt[i-1].y,pt[i].x,pt[i].y,bm,0);
  }
/* draw border from last point to first one */
  draw_edge_(subs,pt[i-1].x,pt[i-1].y,pt[0].x,pt[0].y,bm,0);
    
//     print_bitmap(bm,sw,sh);    
    _bm = bm;
    _bm_w=(sw+7)/8;
    int in = 0,prev_bit=0,bit;
    direct_draw dd(subs);
/* scan bitmap and fill inner space of polygon*/
    for (y=0;y<sh;y++){
        in=0;prev_bit=0;
        for (x=0;x<sw;x++){ 
            bit=getbit(x,y);
            if (bit!=prev_bit){
                if (!bit){
                    if (in&1)
                       in++;
                    else{    
                        int _x;
                        for (_x=x+1;_x<sw;_x++){
                            if (_x&7==0){ 
                                if (!getbyte(_x,y)){
                                    _x+=7;
                                    continue;
                                }else{
                                    in++;
                                    break;
                                }
                            }
                            if (getbit(_x,y)){ 
                                in++;
                                break;
                            }
                        }
                    }
                }
            }
            if (in&1){
                if (!bit){
                  //  if (y && getbit(x,y-1)){
                        dd.setpixel(x,y);
                        setbit(x,y);
//                    }else
//                        in--;
                }
                /*else clearbit(x,y);*/
            }/*else clearbit(x,y);*/
            prev_bit=bit;
        }
    }
    dd.end_draw();
//    print_bitmap(bm,sw,sh); 
    _TRACE("XOR w %d h %d\n",sw,sh);     

/* release bitmask and subsurface*/
    Xfree(bm);
  release_gc(subs);
}
