//
// "$Id: xutils.cxx 16 2007-03-09 13:23:55Z nikego $"
//
// DirectFB-specific code for the Fast Light Tool Kit (FLTK).
//
// Copyright 1998-2005 by Bill Spitzak and others.
// Written by Nikita Egorov (nikego@gmail.com)
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Library General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Library General Public License for more details.
//
// You should have received a copy of the GNU Library General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA.
//
// Please report all bugs and problems on the following page:
//
//     http://www.fltk.org/str.php
//
// This file contains colormap-specific code for fltk which is always linked
// in.  Search other files for "DIRECTFB" or filenames ending in _directfb.cxx
// for other system-specific code.
//
 
#include <config.h>

#ifdef DIRECTFB
#include "FL/x.H"
#include <directfb_version.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>

static void
strlwr(char* buf){
  while(*buf){
    *buf = __tolower(*buf); 
    buf++;
  }
}
#define sqr(a) ((a)*(a))

/*******************
 * XListFonts()
 * XFreeFontNames()
 * XSetFontPath()
 * XGetFontPath()
 * XFreeFontPath()
*/ 	
#include "_listfonts.cxx"
/*******************
 * XLoadFont
*/
#include "_loadfont.cxx"
/*******************
 * XDrawLine()
 * XDrawLines()
 * XDrawRectangle()
 * XFillRectangle()
 * work with non ascii font charset. cyrillic added only...
 * draw_string()
 * _GetStringWidth()
 * implementation of class direct_draw
 * ---- experimental smooth drawing functions
 * draw_smooth_line()
 * SmoothPoly()
 * fl_filter_offscreen()
 * fl_filter_widget()
*/
#include "_draw.cxx"
/********************
 * GdEllipse()
 * GdArc()
 * GdFillPoly()   
 */ 
#include "_arcpoly.cxx"
/*********************
 * XSetForeground()
 * XFreeColors()
 * XFlushColormap() - my function
 * XAllocColor()
 * XCreateColormap()
 * XFreeColormap()
 * XQueryColor()
 * XQueryColors()
*/
#include "_colormap.cxx"
/**********************
 * XPutImage()
 * XCreateBitmapFromData()
 * XDestroyImage()
 * XCreatePixmapCursor()
 * XRecolorCursor()
 * XCreateFontCursor()
 * XDefineCursor()
 * XFreeCursor()  
 * fl_create_alphamask()
 * SetColor* functions group
*/
#include "_image.cxx"

/**********************/
 
typedef struct __gc_item__{
  GC                  gc;
  struct __gc_item__  *next;
}gc_item;

static gc_item   list_head  = {0};
static int       gc_cnt     =  0,
                 win_cnt    =  0;

static inline void
push_gc(GC gc)
{
    gc_item *item   = (gc_item*)Xcalloc(sizeof(gc_item), 1);
    item->next      = list_head.next;
    item->gc        = gc;
    list_head.next  = item;
    gc_cnt++;
}

static inline int
remove_gc(GC gc)
{
  if (gc_cnt > 0){ 
    gc_item *citem = list_head.next;
    gc_item *pitem = &list_head;
    for (;citem;citem=citem->next){
      if (citem->gc == gc){
         pitem->next = citem->next;
         Xfree(citem);
         gc_cnt--;
         return 0;             
      }
      pitem = citem;
    }
  }
  return -1;
}

void 
for_each_gc( int (cb)(GC gc,void* param),void* param)
{
  for (gc_item *citem = list_head.next; citem ; citem=citem->next){
    if (cb(citem->gc,param))
      break;
  }
}

int
print_gc(GC gc,void*)
{
  int w,h;
  get_size(gc, &w, &h);
  D_INFO("width = %d\theight = %d\tpixel format %s\n",\
      w, h, pixelformat_name(get_pixel_format(gc)));
  return 0;
}

void
print_gc_list()
{
  D_INFO("list of surfaces:\n");
  for_each_gc(print_gc,0);
}

void
free_gc_list()
{
  gc_item *pitem = &list_head;
  for (gc_item *citem = pitem->next; citem; citem = pitem->next){
    release(citem->gc->surface);
    Xfree(citem->gc);
    pitem->next = citem->next;
    Xfree(citem);
    gc_cnt--;
  }
}

GC
create_gc(const DFBSurfaceDescription* desc)
{
  if (fl_dfb){
    GC gc = (GC)Xcalloc(sizeof(surface_context),1);
    D_ASSERT( gc );
    DFBCHECK(fl_dfb->CreateSurface(fl_dfb, desc, &gc->surface));
    if (gc->surface){
      push_gc(gc);
      if(!(desc->flags & DSDESC_PREALLOCATED))
        //clear_gc(gc,0,0,0,255);
        clear_gc(gc,0,0,0,0);
      return gc;
    }else
      Xfree( gc );
  }
  return 0;
}

void
release_gc(GC gc)
{
  if (remove_gc(gc) && fl_dfb)
    D_WARN("no such surface %p", gc);
  else
  {
    /*try*/
    {
      if (fl_dfb){
        release(gc->surface);
        Xfree(gc);
      }
    }
    /*catch(...)
    {
      D_WARN("Oops!");
    }*/
  }
  return;
}

GC
create_mem_gc(int w, int h, int pf/*=-1*/)
{
  DFBSurfaceDescription desc;
  desc.caps      = DSCAPS_SYSTEMONLY; 
  desc.flags     =(DFBSurfaceDescriptionFlags)(DSDESC_CAPS|DSDESC_WIDTH|DSDESC_HEIGHT|DSDESC_PIXELFORMAT);
  desc.height    = h;
  desc.width     = w;
    
  if (pf == -1)
    desc.pixelformat = (DFBSurfacePixelFormat)pixel_format();
  else
    desc.pixelformat = (DFBSurfacePixelFormat)pf;
    
  GC gc = create_gc(&desc);
  return gc;
}

GC
create_gc(FL_DFB_Window* window)
{
  GC gc = 0;
  Fl_Window* w1 =  fl_find(window);
  if (w1){
     CreatedWindow* w2 = CreatedWindow::i(w1);
     if (w2)  
       gc = w2->gc;
  }
  if (!gc){
    gc = (GC)Xcalloc(sizeof(surface_context),1);
    window->getSurface(&gc->surface);
    D_DEBUG("gc was created!\n");
  }else{
      gc->surface->AddRef(gc->surface);
  }
  push_gc(gc);
  return gc;
}

GC
create_gc(GC parent_gc, const DFBRectangle *rect)
{
  GC gc = (GC)Xcalloc(sizeof(surface_context),1);
  if (DFB_OK == parent_gc->surface->GetSubSurface(parent_gc->surface, rect, &gc->surface)){
	push_gc(gc);
  }
  return gc;
}

GC
create_gc(int caps, int flags, int pixelformat, int width, int height)
{
  DFBSurfaceDescription desc;
  desc.caps         = (DFBSurfaceCapabilities)caps;
  desc.flags        = (DFBSurfaceDescriptionFlags)flags;
  desc.pixelformat  = (DFBSurfacePixelFormat)pixelformat;
  desc.height       = height;
  desc.width        = width;

  return create_gc(&desc);
}

GC
create_gc(IDirectFBDisplayLayer* layer)
{
    GC gc = (GC)Xcalloc(sizeof(surface_context), 1);
    DFBCHECK(layer->GetSurface(layer, &gc->surface));
    push_gc(gc);
    return gc;
}

Pixmap
create_pixmap(int w ,int h, int pf/* = -1*/)
{
  return (Pixmap)create_mem_gc(w,h,pf);
}

void
release_window(FL_DFB_Window *window)
{
printf(" release window %p\n", window);
  if (window) {
  	delete window;
  	win_cnt--;
  }
}

#if 0
IDirectFBWindow* 
create_window(const DFBWindowDescription* desc)
{
  if (display_layer){
    win_cnt++;
    IDirectFBWindow* ret_win = 0;
    printf("CreateWindow {%d, %d, %d, %d} win_cnt = %d\n", desc->posx, desc->posy, desc->width, desc->height, win_cnt);
    DFBResult err = display_layer->CreateWindow(display_layer, desc, &ret_win);
    if (DFB_OK == err){
      return ret_win;
    }else{
      D_WARN("Cant create the window ! %s", DirectFBErrorString(err));
      return 0;
    }
  }else
    D_WARN("Display layer is absent!\n");
  return 0;
}

//FIXME in fact, I can't use GetWindow()
//I receive an other interface instead gotten one by CreateWindow()
IDirectFBWindow*
get_window(DFBWindowID xid)
{
  if (display_layer){
    IDirectFBWindow* ret_win;
    if (DFB_OK == display_layer->GetWindow(display_layer,xid,&ret_win)){
      return ret_win;
    }
  }
  return 0;
}
#endif

Region
XCreateRegion()
{
  Region r = (Region)Xmalloc(sizeof(DFBRegion));
  r->x1=r->y1=1;
  r->x2=r->y2=0;
  return r;
}
 
void
XDestroyRegion(Region r) { 
  Xfree(r);
}

int
XRectInRegion(Region r,int x,int y,int w,int h) 
{
  if (dfb_region_intersects(r, DFB_REGION_VALS_FROM_RECTANGLE_VALS(x,y,w+1,h+1)))
    return 2;
  
  return 0;  
}

Region 
XRectangleRegion(int x, int y, int w, int h) {
  Region r = XCreateRegion();
  *r = DFB_REGION_INIT_FROM_RECTANGLE_VALS(x,y,w+1,h+1);
  return r;
}

void
XClipBox(Region r, XRectangle* rect) {
    dfb_rectangle_from_region((DFBRectangle*)rect , r);
    rect->w--;
    rect->h--;
}

const int max_regions = 32;

void 
update_region(GC gc, Region r)
{
  int caps = get_caps(gc);
  DFBRegion region;
  
  if (!((caps & DSCAPS_DOUBLE) || (caps & DSCAPS_TRIPLE)))
    return;
  
  if (!r){
    int w,h;
    get_size(gc, &w, &h);
    region = DFB_REGION_INIT_FROM_RECTANGLE_VALS(0,0,w,h);
    r = &region;
  }
  r->x2++;r->y2++;//FIXME
  if (r->x2 <= r->x1 || r->y2 <= r->y1)
      return;
  //D_DEBUG("update_region(%p,{%d,%d,%d,%d})\n",surface,r->x1,r->y1,r->x2,r->y2);

  if (gc->updates.regions == 0) {
    DFBRegion *regions = (DFBRegion *)Xmalloc(sizeof(DFBRegion)*max_regions);
    dfb_updates_init(&gc->updates, regions, max_regions);
  }
  dfb_updates_add(&gc->updates, r);
}

void
update_region(GC gc,int x1,int y1,int x2,int y2)
{
	if (x1>x2){int tmp=x1;x1=x2;x2=tmp;}
	if (y1>y2){int tmp=y1;y1=y2;y2=tmp;}
	
	DFBRegion reg = {x1,y1,x2,y2};	
    
	update_region(gc, &reg);
}

void 
updates_reset(GC gc)
{
  if (gc && gc->updates.regions)
    dfb_updates_reset(&gc->updates);
  else
    D_WARN("gc is NULL!");
}

void 
remove_updates(GC gc)
{
  if (gc && gc->updates.regions) {
    dfb_updates_reset(&gc->updates);
    Xfree(gc->updates.regions);
    gc->updates.regions = 0;
  }else
    D_WARN("gc is NULL!");
}

void 
flip_updates(GC gc)
{
  int total,bounding;
  if (gc){
    DFBUpdates* updates = &gc->updates;
    if (updates->num_regions){
      dfb_updates_stat(updates, &total, &bounding);
      if (bounding <= total)
        fl_flip(gc, &updates->bounding);
      else{
        int j = 0;
        for (j=0; j<updates->num_regions; j++)
          fl_flip(gc, updates->regions+j);
      }
      dfb_updates_reset(updates);
    }
  }else
    D_WARN("gc is NULL!");
}

void
XSetClipMask(Display* display, GC gc, Pixmap pixmap)
{
  if (pixmap == 0){
    if (gc)
      XSetRegion(display, gc, 0);
  }
  if (gc)
    gc->clip_mask = (surface_context*)pixmap;
}

//set region for clip
void 
XSetRegion(Display* display, GC surface, Region r, int _set_clip)
{
  if (!surface)
    return;

  if (!r){
    static DFBRegion region;
    int w,h;
    get_size(surface, &w, &h);
    region = DFB_REGION_INIT_FROM_RECTANGLE_VALS(0,0,w+1,h+1);
    r = &region;
  }
  //DirectFB's function to clip the region
  if (_set_clip)
    set_clip(surface, r);
}

int
XWindowEvent(Display* display, Window w, long event_mask, XEvent * ev)
{
    /* FLTK hack*/
    if (event_mask == ExposureMask)
        ev->type = NoExpose;

    return 0;
}

int
XIntersectRegion(Region r1, Region r2, Region newReg)
{
  *newReg = r1 ? *r1 : *r2;
  return dfb_region_region_intersect(newReg, r2 ? r2 : r1);
}

int
XSetClipOrigin(Display* display, GC gc, int clip_x_origin, int clip_y_origin){
  _TRACE("XSetClipOrigin() is called\n");
  return 1;
}

int
XUnionRectWithRegion(XRectangle* rect, Region src_reg, Region dst_reg)
{
  static DFBRegion region;
  dfb_region_from_rectangle(&region, rect);
  region.x2++;region.y2++;
  dfb_region_region_union(&region, src_reg);
  *dst_reg = region;
  return 1;
}

int
XSubtractRegion(Region sra,Region srb,Region dr_return) {
    D_INFO("XSubtractRegion() is called\n");
    return 0;
}

int
XParseGeometry(const char *parsestring, int *x_return, int 
	*y_return, unsigned int *width_return, unsigned int *height_return){
	return 0;	
}

void
XDestroyWindow(Display* display , dfbWindow xid)
{
  //ungrab_pointer(xid);
  //DFBCHECK(xid->Destroy(xid));
  release_window(xid);
}

int
XIconifyWindow(Display*, Window xid, void*){
	Fl_Window* win = fl_find(xid);
	CreatedWindow::i(win)->set_minimize();
	return 0;
}

//extern int handle(const DFBWindowEvent& thisevent);
#if 0
int
XMapWindow(Display*, dfbWindow xid)
{
    Fl_Window* win = fl_find(xid);
    CreatedWindow* i = CreatedWindow::i(win);
    if (i){
      if (!i->first_draw){ 
        set_opacity(i->xid, 255);
      }
      if (i->minimize()){
        i->delete_icon();
      }
      i->wait_for_expose = 0;
      redraw();
    }
    return 0;     
}

int
XUnmapWindow(Display* d, dfbWindow xid){
    Fl_Window* win = fl_find(xid);
    CreatedWindow* i = CreatedWindow::i(win);
    if (i){ 
      set_opacity(i->xid, 0);
      i->wait_for_expose = 1;
    }
    return 0;
}

int
XMapRaised(Display* d, dfbWindow xid){
    Fl_Window* win = fl_find(xid);
    CreatedWindow* i = CreatedWindow::i(win);
    //D_INFO("restore: i= %p icon %p win %p\n",i, i->icon, win);
    if (i && i->minimize()){
        i->delete_icon();
        redraw();
    }
    raise_to_top(xid); 
    i->wait_for_expose = 0;
    return 0;
}

GC 
XCreateGC(Display* display,Drawable d,unsigned long valuemask,void* values)
{
  D_ONCE("XCreateGC() not implemented."); 
  return 0;
}

int XBell(Display*, int)
{
    return 0;
}

int
XFreeFont(Display* d, XFont font){
  if (font)
        release(font);
    return 0;    
}
#endif

int
flip_gc(GC gc,void*)
{
  flip_updates(gc);
  return 0;
}

void XFlush(Display* display)
{
  for_each_gc(flip_gc, 0);//FIXME
#if USE_COLORMAP
  // DirectFB's palette function SetEntries() sometimes is very slooow
  // I upload current colormap to the palette when it's needed
  if (have_lut())
    XFlushColormap(fl_display, fl_colormap);
#endif
  extern ulong flip_size_sum;
  if (flip_size_sum){
    _TRACE("flip pixels %lu\n", flip_size_sum);
    flip_size_sum=0;
  }
}

/****************************************************************************
 * Key mapping functions
 */
 
extern DFBInputDeviceID keyboard_id;

static const struct {unsigned short id, fltk; const char* s;} idtab[] = {
	{DIKI_UNKNOWN,0,""},
	{DIKI_A,			'a',		"A"},
    {DIKI_B,			'b',		"B"},
    {DIKI_C,			'c',		"C"},
    {DIKI_D,			'd',		"D"},
    {DIKI_E,			'e',		"E"},
    {DIKI_F,			'f',		"F"},
    {DIKI_G,			'g',		"G"},
    {DIKI_H,			'h',		"H"},
    {DIKI_I,			'i',		"I"},
    {DIKI_J,			'j',		"A"},
    {DIKI_K,			'k',		"A"},
    {DIKI_L,			'l',		"A"},
    {DIKI_M,			'm',		"A"},
    {DIKI_N,			'n',		"A"},
    {DIKI_O,			'o',		"A"},
    {DIKI_P,			'p',		"A"},
    {DIKI_Q,			'q',		"A"},
    {DIKI_R,			'r',		"A"},
    {DIKI_S,			's',		"A"},
    {DIKI_T,			't',		"A"},
    {DIKI_U,			'u',		"A"},
    {DIKI_V,			'v',		"A"},
    {DIKI_W,			'w',		"A"},
    {DIKI_X,			'x',		"A"},
    {DIKI_Y,			'y',		"A"},
    {DIKI_Z,			'z',		"A"},
    {DIKI_0,			'0',		"A"},
    {DIKI_1,			'1',		"A"},
    {DIKI_2,			'2',		"A"},
    {DIKI_3,			'3',		"A"},
    {DIKI_4,			'4',		"A"},
    {DIKI_5,			'5',		"A"},
    {DIKI_6,			'6',		"A"},
    {DIKI_7,			'7',		"A"},
    {DIKI_8,			'8',		"A"},
    {DIKI_9,			'9',		"A"},
	{DIKI_F1,			FL_F+1,		"FL_F+1"},
    {DIKI_F2,			FL_F+2,		"FL_F+2"},
    {DIKI_F3,			FL_F+3,		"FL_F+3"},
    {DIKI_F4,			FL_F+4,		"FL_F+4"},
    {DIKI_F5,			FL_F+5,		"FL_F+5"},
    {DIKI_F6,			FL_F+6,		"FL_F+6"},
    {DIKI_F7,			FL_F+7,		"FL_F+7"},
    {DIKI_F8,			FL_F+8,		"FL_F+8"},
    {DIKI_F9,			FL_F+9,		"FL_F+9"},
    {DIKI_F10,			FL_F+10,	"FL_F+10"},
    {DIKI_F11,			FL_F+11,	"FL_F+11"},
    {DIKI_F12,			FL_F+12,	"FL_F+12"},
    {DIKI_SHIFT_L,		FL_Shift_L,	"FL_Shift_L"},
    {DIKI_SHIFT_R,		FL_Shift_R,	"FL_Shift_R"},
    {DIKI_CONTROL_L,	FL_Control_L,"FL_Control_L"},
    {DIKI_CONTROL_R,	FL_Control_R,"FL_Control_R"},
    {DIKI_ALT_L,		FL_Alt_L,	"FL_Alt_L"},
	{DIKI_ALT_R,		FL_Alt_R,	"FL_Alt_R"},
    //{DIKI_ALTGR,		FL_Alt_R,	"FL_Alt_Gr"},
    {DIKI_META_L,		FL_Meta_L,	"FL_Meta_L"},
    {DIKI_META_R,		FL_Meta_R,	"FL_Meta_R"},
    {DIKI_SUPER_L,		0,			"0"},
    {DIKI_SUPER_R,		FL_Menu,	"0"},
    {DIKI_HYPER_L,		0,			"0"},
    {DIKI_HYPER_R,		0,			"0"},
    {DIKI_CAPS_LOCK,	FL_Caps_Lock,"FL_Caps_Lock"},
    {DIKI_NUM_LOCK,		FL_Num_Lock,"FL_Num_Lock"},
    {DIKI_SCROLL_LOCK,	FL_Scroll_Lock,"FL_Scroll_Lock"},
    {DIKI_ESCAPE,		FL_Escape,	"FL_Escape"},
    {DIKI_LEFT,			FL_Left,	"FL_Left"},
    {DIKI_RIGHT,		FL_Right,	"FL_Right"},
    {DIKI_UP,			FL_Up,		"FL_Up"},
    {DIKI_DOWN,			FL_Down,	"FL_Down"},
    {DIKI_TAB,			FL_Tab,		"FL_Tab"},
    {DIKI_ENTER,		FL_Enter,	"FL_Enter"},
    {DIKI_SPACE,		0x20,		" "},
    {DIKI_BACKSPACE,	FL_BackSpace,"FL_BackSpace"},
    {DIKI_INSERT,		FL_Insert,	"FL_Insert"},
    {DIKI_DELETE,		FL_Delete,	"FL_Delete"},
    {DIKI_HOME,			FL_Home,	"FL_Home"},
    {DIKI_END,			FL_End,		"FL_End"},
    {DIKI_PAGE_UP,		FL_Page_Up,	"FL_Page_Up"},
    {DIKI_PAGE_DOWN,	FL_Page_Down,"FL_Page_Down"},
    {DIKI_PRINT,		FL_Print,	"FL_Print"},
    {DIKI_PAUSE,		FL_Pause,	"FL_Pause"},
    {DIKI_QUOTE_LEFT,   '{',		"{"},
    {DIKI_MINUS_SIGN,   '-',		"-"},
    {DIKI_EQUALS_SIGN,  '=',		"="},
    {DIKI_BRACKET_LEFT, '[',		"["},
    {DIKI_BRACKET_RIGHT,']',		"]"},
    {DIKI_BACKSLASH,    '|',		"|"},
    {DIKI_SEMICOLON,    ';',		";"},
    {DIKI_QUOTE_RIGHT,  '\'',		"\'"},
    {DIKI_COMMA,        ',',		","},
    {DIKI_PERIOD,       '.',		"."},
    {DIKI_SLASH,        '/',		"/"},
    {DIKI_LESS_SIGN,    '<',		"<"},
    {DIKI_KP_DIV,		FL_KP+'/',	""},
    {DIKI_KP_MULT,		FL_KP+'*',	""},
    {DIKI_KP_MINUS,		FL_KP+'-',	""},
    {DIKI_KP_PLUS,		FL_KP+'+',	""},
    {DIKI_KP_ENTER,		FL_KP_Enter,""},
    {DIKI_KP_SPACE,		FL_KP+' ',	""},
    {DIKI_KP_TAB,		FL_KP+'\t',	""},
//  {DIKI_KP_F1,		FL_KP+'/',	""},
//  {DIKI_KP_F2,		FL_KP+'/',	""},
//  {DIKI_KP_F3,		FL_KP+'/',	""},
//  {DIKI_KP_F4,		FL_KP+'/',	""},
    {DIKI_KP_EQUAL,		FL_KP+'=',	""},
    {DIKI_KP_SEPARATOR,	FL_KP+'|',	""},
    {DIKI_KP_DECIMAL,	FL_KP+'.',	"Fl_Delete"},
    {DIKI_KP_0,			FL_KP+'0',	"FL_Insert"},
    {DIKI_KP_1,			FL_KP+'1',	"FL_End"},
    {DIKI_KP_2,			FL_KP+'2',	"FL_Down"},
    {DIKI_KP_3,			FL_KP+'3',	"FL_Page_Down"},
    {DIKI_KP_4,			FL_KP+'4',	"FL_Left"},
    {DIKI_KP_5,			FL_KP+'5',	"FL_KP+5"},
    {DIKI_KP_6,			FL_KP+'6',	"FL_Right"},
    {DIKI_KP_7,			FL_KP+'7',	"FL_Home"},
    {DIKI_KP_8,			FL_KP+'8',	"FL_Up"},
    {DIKI_KP_9,			FL_KP+'9',	"FL_Page_Up"},    {0,0}
};

int
id2fltk(int id) {
	for (int i = 0; idtab[i].id; i++)
		if (idtab[i].id == id){
			return idtab[i].fltk;
		}
	return 0;
} 

int
fltk2id(int sym) {
	for (int i = 0; idtab[i].id; i++)
		if (idtab[i].fltk == sym){
			return idtab[i].id;
		}
	return 0;
} 

bool fl_FLEventToDFBEvent(int e, DFBEvent& dfbevent)
{
#if 1 //filter all event produced by FLTK.FIXME
    extern DFBEvent* fl_dfb_event;
    
    if (!fl_dfb_event || fl_dfb_event->window.clazz != DFEC_WINDOW)
        return false;
    if (e != FL_PUSH 
        && e != FL_RELEASE 
        && e != FL_MOVE
        && e != FL_DRAG
        && e != FL_SHORTCUT 
        && e != FL_KEYDOWN 
        && e != FL_KEYUP)
        return false;
    if (e == FL_PUSH && fl_dfb_event->window.type != DWET_BUTTONDOWN)
        return false;
    if (e == FL_RELEASE && fl_dfb_event->window.type != DWET_BUTTONUP)
        return false;
    if ((e == FL_MOVE || e == FL_DRAG) && fl_dfb_event->window.type != DWET_MOTION)
        return false;
    if ((e == FL_KEYDOWN || e == FL_SHORTCUT) && fl_dfb_event->window.type != DWET_KEYDOWN)
        return false;
    if (e == FL_KEYUP && fl_dfb_event->window.type != DWET_KEYUP)
        return false;
    dfbevent = *fl_dfb_event;
    
    if (e == FL_PUSH 
        || e == FL_RELEASE 
        || e == FL_MOVE 
        || e == FL_DRAG) {
        dfbevent.window.cx = Fl::event_x();
        dfbevent.window.cy = Fl::event_y();
    }
    return true;
#else
    memset(&dfbevent, 0, sizeof(DFBEvent));   
    dfbevent.window.clazz = DFEC_WINDOW;
    dfbevent.window.window_id = 0; //FIXME    
    
    if (Fl::event_shift())
        dfbevent.window.modifiers = (DFBInputDeviceModifierMask)(dfbevent.window.modifiers | DIMM_SHIFT);
    if (Fl::event_ctrl())
        dfbevent.window.modifiers = (DFBInputDeviceModifierMask)(dfbevent.window.modifiers | DIMM_CONTROL);
    if (Fl::event_alt())
        dfbevent.window.modifiers = (DFBInputDeviceModifierMask)(dfbevent.window.modifiers | DIMM_ALT);
    if (Fl::event_state(FL_META))
        dfbevent.window.modifiers = (DFBInputDeviceModifierMask)(dfbevent.window.modifiers | DIMM_META);

    if (Fl::event_button1())
        dfbevent.window.buttons = (DFBInputDeviceButtonMask)(dfbevent.window.buttons | DIBM_LEFT);
    if (Fl::event_button2())
        dfbevent.window.buttons = (DFBInputDeviceButtonMask)(dfbevent.window.buttons | DIBM_MIDDLE);
    if (Fl::event_button3())
        dfbevent.window.buttons = (DFBInputDeviceButtonMask)(dfbevent.window.buttons | DIBM_RIGHT);

    if (Fl::event_state(FL_NUM_LOCK))
        dfbevent.window.locks = (DFBInputDeviceLockState)(dfbevent.window.locks | DILS_NUM);
    if (Fl::event_state(FL_CAPS_LOCK))
        dfbevent.window.locks = (DFBInputDeviceLockState)(dfbevent.window.locks | DILS_CAPS);
    if (Fl::event_state(FL_SCROLL_LOCK))
        dfbevent.window.locks = (DFBInputDeviceLockState)(dfbevent.window.locks | DILS_SCROLL);

    gettimeofday(&dfbevent.window.timestamp, NULL);
    
    switch(e) {
        case FL_PUSH:            
        case FL_RELEASE:
        case FL_MOVE:
        case FL_DRAG:
            dfbevent.window.type = (e == FL_PUSH) ? DWET_BUTTONDOWN : 
                                   (e == FL_RELEASE) ? DWET_BUTTONUP : DWET_MOTION;
            dfbevent.window.cx = Fl::event_x();
            dfbevent.window.cy = Fl::event_y();
            return true;
        case FL_SHORTCUT:
        case FL_KEYDOWN:
        case FL_KEYUP:
            dfbevent.window.type = (e == FL_KEYUP) ? DWET_KEYUP : DWET_KEYDOWN;
            dfbevent.window.key_id = (DFBInputDeviceKeyIdentifier)fltk2id(Fl::event_key());
            dfbevent.window.key_code = -1;
            if (Fl::event_length() && Fl::event_text())
                dfbevent.window.key_symbol = (DFBInputDeviceKeySymbol)(Fl::event_text()[0]);
            return true;
        default:
            return false;
    }
#endif
}

int
XKeysymToKeycode(Display* d, int ks)
{
	DFBInputDeviceKeymapEntry entry;
	IDirectFBInputDevice* kbd;
	int code = 0,id = fltk2id(ks);
	fl_dfb->GetInputDevice(fl_dfb, keyboard_id, &kbd);
	for (int i=0; i< 255;i++){
		kbd->GetKeymapEntry(kbd, i, &entry);
		if (id!=0){
			if (entry.identifier==id){
				code = entry.code;
				break;
			}
		}else{
			for (int j = 0; j <= DIKSI_LAST; j++){
				if (entry.symbols[j]==ks){
					code = entry.code;
					break;
				}
			}
		}
	}
	release(kbd);
	return code; 
}

#endif
